/*
  Copyright 2012 Computer Vision Lab,
  Ecole Polytechnique Federale de Lausanne (EPFL), Switzerland.
  All rights reserved.

  Authors: Tomasz Trzcinski, Eray Molla, and Vincent Lepetit

  This file is part of the DBRIEF_demo software.

  DBRIEF_demo is  free software; you can redistribute  it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free  Software Foundation; either  version 2 of the  License, or
  (at your option) any later version.

  DBRIEF_demo is  distributed in the hope  that it will  be useful, but
  WITHOUT  ANY   WARRANTY;  without  even  the   implied  warranty  of
  MERCHANTABILITY  or FITNESS FOR  A PARTICULAR  PURPOSE. See  the GNU
  General Public License for more details.

  You should  have received a copy  of the GNU  General Public License
  along  with DBRIEF_demo;  if  not,  write  to   the  Free  Software
  Foundation,  Inc.,  51  Franklin  Street, Fifth  Floor,  Boston,  MA
  02110-1301, USA
*/

#include "Dbrief.h"

#include <cv.h>
#include <highgui.h>
#include <cvaux.h>
#include <ctime>
#include <iostream>

#include <omp.h>


using namespace std;
using namespace CVLAB;

Dbrief::Dbrief(void)
{
  boxSmoothedImage = cvCreateMat(1, 1, CV_32FC1);
}

Dbrief::~Dbrief(void)
{
 cvReleaseMat(&boxSmoothedImage);
}

void Dbrief::getDbriefDescriptors(vector< bitset<DESC_LEN> >& descriptors,
                              const vector<cv::KeyPoint>& kpts,
                               IplImage* img)
{
 // Make sure that input image contains only one color channel:
 assert(img->nChannels == 1);

 // Check whether all the keypoints are inside the subimage:
 assert(validateKeypoints(kpts, img->width, img->height));

 // If memory allocated in 'descriptors' is not enough, then resize it:
 descriptors.resize(kpts.size());

 // Allocate memory for the box smoothed image
 allocateBoxSmoothedImage(img);

 // Calculate the box smoothed image:
 cvSmooth(img, boxSmoothedImage, CV_BLUR_NO_SCALE, KERNEL_SIZE, KERNEL_SIZE);

 // Iterate over keypoints:
 //#pragma omp parallel for shared(img)
 for (unsigned int i = 0; i < kpts.size(); ++i)
  getDbriefDescriptor(descriptors[i], kpts[i], img);
}

bool Dbrief::isKeypointInsideSubImage(const cv::KeyPoint& kpt, const int width, const int height)
{
  return
    SUBIMAGE_LEFT <= kpt.pt.x  &&  kpt.pt.x < SUBIMAGE_RIGHT(width) &&
    SUBIMAGE_TOP  <= kpt.pt.y  &&  kpt.pt.y < SUBIMAGE_BOTTOM(height);
}

bool Dbrief::validateKeypoints(const vector<cv::KeyPoint>& kpts, int im_w, int im_h)
{
  for (unsigned int i = 0; i < kpts.size(); ++i)
    if ( !isKeypointInsideSubImage(kpts[i], im_w, im_h) )
      return false;

  return true;
}

void Dbrief::allocateBoxSmoothedImage(const IplImage* img)
{
  const int im_w_1 = img->width, im_h_1 = img->height;

  if (im_w_1 != boxSmoothedImage->width && im_h_1 != boxSmoothedImage->height) {
    cvReleaseMat(&boxSmoothedImage);
    boxSmoothedImage = cvCreateMat(im_h_1, im_w_1, CV_32FC1);
  }
}

void Dbrief::getDbriefDescriptor(bitset<DESC_LEN>& desc, cv::KeyPoint kpt, IplImage* img)
{
  //  Calculate the width step of the boxSmoothedImage image
  int inWS = boxSmoothedImage->step / CV_ELEM_SIZE(boxSmoothedImage->type);
  //  Hold the pointer of the data part of boxSmoothedImage image matrix with "iD"
  float* iD = boxSmoothedImage->data.fl;
  // Hold the pointer to the top left corner of the patch to be analysed
  int X = kpt.pt.x - HALF_PATCH_SIZE,
      Y = kpt.pt.y - HALF_PATCH_SIZE;
  cv::Point2i point;
  float result;

 /***********************************************
                  Projection: 1
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+1,X+1, inWS) * (0.16649845f);
 result += GET_MATRIX_DATA(iD, Y+32,X+2, inWS) * (0.11555521f);
 result += GET_MATRIX_DATA(iD, Y+1,X+3, inWS) * (0.07356358f);
 result += GET_MATRIX_DATA(iD, Y+15,X+5, inWS) * (-0.03134872f);
 result += GET_MATRIX_DATA(iD, Y+15,X+6, inWS) * (-0.03080348f);
 result += GET_MATRIX_DATA(iD, Y+21,X+6, inWS) * (-0.08975087f);
 result += GET_MATRIX_DATA(iD, Y+2,X+7, inWS) * (0.07508487f);
 result += GET_MATRIX_DATA(iD, Y+11,X+8, inWS) * (-0.01763125f);
 result += GET_MATRIX_DATA(iD, Y+28,X+8, inWS) * (0.12821266f);
 result += GET_MATRIX_DATA(iD, Y+13,X+9, inWS) * (-0.02932138f);
 result += GET_MATRIX_DATA(iD, Y+27,X+9, inWS) * (-0.08967076f);
 result += GET_MATRIX_DATA(iD, Y+29,X+9, inWS) * (-0.02604919f);
 result += GET_MATRIX_DATA(iD, Y+17,X+10, inWS) * (-0.28344749f);
 result += GET_MATRIX_DATA(iD, Y+32,X+10, inWS) * (0.00380377f);
 result += GET_MATRIX_DATA(iD, Y+13,X+11, inWS) * (-0.21699047f);
 result += GET_MATRIX_DATA(iD, Y+14,X+11, inWS) * (-0.03556371f);
 result += GET_MATRIX_DATA(iD, Y+23,X+11, inWS) * (-0.11133723f);
 result += GET_MATRIX_DATA(iD, Y+26,X+11, inWS) * (-0.00829682f);
 result += GET_MATRIX_DATA(iD, Y+19,X+12, inWS) * (-0.16360198f);
 result += GET_MATRIX_DATA(iD, Y+20,X+12, inWS) * (-0.10134142f);
 result += GET_MATRIX_DATA(iD, Y+32,X+12, inWS) * (0.05540031f);
 result += GET_MATRIX_DATA(iD, Y+9,X+13, inWS) * (-0.19363108f);
 result += GET_MATRIX_DATA(iD, Y+13,X+13, inWS) * (-0.12239592f);
 result += GET_MATRIX_DATA(iD, Y+15,X+13, inWS) * (-0.24540921f);
 result += GET_MATRIX_DATA(iD, Y+19,X+13, inWS) * (-0.07538517f);
 result += GET_MATRIX_DATA(iD, Y+22,X+13, inWS) * (-0.02055112f);
 result += GET_MATRIX_DATA(iD, Y+12,X+14, inWS) * (0.07766059f);
 result += GET_MATRIX_DATA(iD, Y+14,X+14, inWS) * (-0.00126232f);
 result += GET_MATRIX_DATA(iD, Y+17,X+14, inWS) * (-0.53326587f);
 result += GET_MATRIX_DATA(iD, Y+11,X+15, inWS) * (-0.15860261f);
 result += GET_MATRIX_DATA(iD, Y+14,X+15, inWS) * (-0.76122419f);
 result += GET_MATRIX_DATA(iD, Y+17,X+15, inWS) * (-0.10178697f);
 result += GET_MATRIX_DATA(iD, Y+19,X+15, inWS) * (-0.20611716f);
 result += GET_MATRIX_DATA(iD, Y+20,X+15, inWS) * (-0.24212381f);
 result += GET_MATRIX_DATA(iD, Y+2,X+16, inWS) * (0.01115456f);
 result += GET_MATRIX_DATA(iD, Y+6,X+16, inWS) * (-0.06620357f);
 result += GET_MATRIX_DATA(iD, Y+10,X+16, inWS) * (0.25576031f);
 result += GET_MATRIX_DATA(iD, Y+11,X+17, inWS) * (-0.54820445f);
 result += GET_MATRIX_DATA(iD, Y+17,X+17, inWS) * (0.79227281f);
 result += GET_MATRIX_DATA(iD, Y+22,X+17, inWS) * (-0.33684689f);
 result += GET_MATRIX_DATA(iD, Y+17,X+18, inWS) * (0.03094824f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (0.40243482f);
 result += GET_MATRIX_DATA(iD, Y+27,X+18, inWS) * (-0.02858646f);
 result += GET_MATRIX_DATA(iD, Y+29,X+19, inWS) * (-0.00512908f);
 result += GET_MATRIX_DATA(iD, Y+5,X+20, inWS) * (-0.06576423f);
 result += GET_MATRIX_DATA(iD, Y+6,X+20, inWS) * (-0.05354734f);
 result += GET_MATRIX_DATA(iD, Y+7,X+20, inWS) * (-0.03675724f);
 result += GET_MATRIX_DATA(iD, Y+27,X+20, inWS) * (-0.00418666f);
 result += GET_MATRIX_DATA(iD, Y+13,X+21, inWS) * (-0.02592852f);
 result += GET_MATRIX_DATA(iD, Y+18,X+21, inWS) * (0.24830766f);
 result += GET_MATRIX_DATA(iD, Y+22,X+21, inWS) * (0.00654599f);
 result += GET_MATRIX_DATA(iD, Y+13,X+22, inWS) * (-0.03317427f);
 result += GET_MATRIX_DATA(iD, Y+25,X+22, inWS) * (-0.02324759f);
 result += GET_MATRIX_DATA(iD, Y+11,X+24, inWS) * (-0.04599604f);
 result += GET_MATRIX_DATA(iD, Y+23,X+24, inWS) * (-0.05936127f);
 result += GET_MATRIX_DATA(iD, Y+1,X+26, inWS) * (0.08725166f);
 result += GET_MATRIX_DATA(iD, Y+23,X+26, inWS) * (-0.03869978f);
 result += GET_MATRIX_DATA(iD, Y+32,X+27, inWS) * (0.02099043f);
 result += GET_MATRIX_DATA(iD, Y+12,X+28, inWS) * (-0.13404850f);
 result += GET_MATRIX_DATA(iD, Y+19,X+28, inWS) * (-0.03797154f);
 result += GET_MATRIX_DATA(iD, Y+2,X+29, inWS) * (-0.23796032f);
 result += GET_MATRIX_DATA(iD, Y+1,X+30, inWS) * (0.23972543f);
 result += GET_MATRIX_DATA(iD, Y+25,X+30, inWS) * (-0.01397713f);
 result += GET_MATRIX_DATA(iD, Y+1,X+32, inWS) * (0.16828351f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00273870f);
desc[0] = (result > (-9231.34f));

 /***********************************************
                  Projection: 2
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+9,X+1, inWS) * (0.00592549f);
 result += GET_MATRIX_DATA(iD, Y+15,X+1, inWS) * (0.03712640f);
 result += GET_MATRIX_DATA(iD, Y+6,X+3, inWS) * (0.01098857f);
 result += GET_MATRIX_DATA(iD, Y+13,X+3, inWS) * (0.01450933f);
 result += GET_MATRIX_DATA(iD, Y+15,X+3, inWS) * (0.00299020f);
 result += GET_MATRIX_DATA(iD, Y+19,X+3, inWS) * (0.00756036f);
 result += GET_MATRIX_DATA(iD, Y+25,X+3, inWS) * (0.03198459f);
 result += GET_MATRIX_DATA(iD, Y+19,X+4, inWS) * (0.01914929f);
 result += GET_MATRIX_DATA(iD, Y+8,X+5, inWS) * (0.03857332f);
 result += GET_MATRIX_DATA(iD, Y+10,X+5, inWS) * (0.00507683f);
 result += GET_MATRIX_DATA(iD, Y+11,X+5, inWS) * (0.03925057f);
 result += GET_MATRIX_DATA(iD, Y+3,X+6, inWS) * (0.00153810f);
 result += GET_MATRIX_DATA(iD, Y+17,X+6, inWS) * (0.04242534f);
 result += GET_MATRIX_DATA(iD, Y+21,X+6, inWS) * (0.09040241f);
 result += GET_MATRIX_DATA(iD, Y+27,X+6, inWS) * (0.01248976f);
 result += GET_MATRIX_DATA(iD, Y+13,X+7, inWS) * (0.03741935f);
 result += GET_MATRIX_DATA(iD, Y+16,X+7, inWS) * (0.01474961f);
 result += GET_MATRIX_DATA(iD, Y+23,X+8, inWS) * (0.02267983f);
 result += GET_MATRIX_DATA(iD, Y+29,X+8, inWS) * (0.03532362f);
 result += GET_MATRIX_DATA(iD, Y+13,X+9, inWS) * (0.01692446f);
 result += GET_MATRIX_DATA(iD, Y+17,X+9, inWS) * (0.06630675f);
 result += GET_MATRIX_DATA(iD, Y+7,X+10, inWS) * (0.05029527f);
 result += GET_MATRIX_DATA(iD, Y+11,X+10, inWS) * (0.04127525f);
 result += GET_MATRIX_DATA(iD, Y+22,X+10, inWS) * (0.01326443f);
 result += GET_MATRIX_DATA(iD, Y+9,X+11, inWS) * (0.00994800f);
 result += GET_MATRIX_DATA(iD, Y+11,X+11, inWS) * (0.00586611f);
 result += GET_MATRIX_DATA(iD, Y+25,X+11, inWS) * (0.00264126f);
 result += GET_MATRIX_DATA(iD, Y+18,X+13, inWS) * (0.01586773f);
 result += GET_MATRIX_DATA(iD, Y+24,X+13, inWS) * (0.04002828f);
 result += GET_MATRIX_DATA(iD, Y+16,X+15, inWS) * (0.12015208f);
 result += GET_MATRIX_DATA(iD, Y+17,X+15, inWS) * (0.00050466f);
 result += GET_MATRIX_DATA(iD, Y+16,X+17, inWS) * (0.05005587f);
 result += GET_MATRIX_DATA(iD, Y+18,X+17, inWS) * (0.08671221f);
 result += GET_MATRIX_DATA(iD, Y+23,X+20, inWS) * (-0.00373608f);
 result += GET_MATRIX_DATA(iD, Y+25,X+20, inWS) * (-0.07878207f);
 result += GET_MATRIX_DATA(iD, Y+6,X+21, inWS) * (-0.02152568f);
 result += GET_MATRIX_DATA(iD, Y+9,X+21, inWS) * (-0.03441605f);
 result += GET_MATRIX_DATA(iD, Y+16,X+21, inWS) * (0.01658224f);
 result += GET_MATRIX_DATA(iD, Y+18,X+21, inWS) * (0.03917277f);
 result += GET_MATRIX_DATA(iD, Y+11,X+23, inWS) * (-0.02475726f);
 result += GET_MATRIX_DATA(iD, Y+31,X+23, inWS) * (-0.02336995f);
 result += GET_MATRIX_DATA(iD, Y+9,X+24, inWS) * (-0.03497746f);
 result += GET_MATRIX_DATA(iD, Y+23,X+24, inWS) * (-0.05346992f);
 result += GET_MATRIX_DATA(iD, Y+1,X+25, inWS) * (-0.00913155f);
 result += GET_MATRIX_DATA(iD, Y+14,X+25, inWS) * (-0.01786428f);
 result += GET_MATRIX_DATA(iD, Y+12,X+26, inWS) * (-0.00455171f);
 result += GET_MATRIX_DATA(iD, Y+21,X+26, inWS) * (-0.02484205f);
 result += GET_MATRIX_DATA(iD, Y+6,X+27, inWS) * (-0.00634494f);
 result += GET_MATRIX_DATA(iD, Y+8,X+27, inWS) * (-0.02251113f);
 result += GET_MATRIX_DATA(iD, Y+22,X+27, inWS) * (-0.02171910f);
 result += GET_MATRIX_DATA(iD, Y+23,X+27, inWS) * (-0.01885075f);
 result += GET_MATRIX_DATA(iD, Y+27,X+27, inWS) * (-0.01376151f);
 result += GET_MATRIX_DATA(iD, Y+30,X+27, inWS) * (-0.00865876f);
 result += GET_MATRIX_DATA(iD, Y+31,X+27, inWS) * (-0.01426335f);
 result += GET_MATRIX_DATA(iD, Y+19,X+28, inWS) * (-0.00613425f);
 result += GET_MATRIX_DATA(iD, Y+3,X+30, inWS) * (-0.03808357f);
 result += GET_MATRIX_DATA(iD, Y+10,X+30, inWS) * (-0.07337978f);
 result += GET_MATRIX_DATA(iD, Y+13,X+30, inWS) * (-0.01057705f);
 result += GET_MATRIX_DATA(iD, Y+16,X+30, inWS) * (-0.10765289f);
 result += GET_MATRIX_DATA(iD, Y+19,X+30, inWS) * (-0.00112509f);
 result += GET_MATRIX_DATA(iD, Y+21,X+30, inWS) * (-0.06285825f);
 result += GET_MATRIX_DATA(iD, Y+31,X+30, inWS) * (-0.03361454f);
 result += GET_MATRIX_DATA(iD, Y+25,X+31, inWS) * (-0.08592733f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00449087f);
desc[1] = (result > (-336.06f));

 /***********************************************
                  Projection: 3
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+24,X+1, inWS) * (-0.02341080f);
 result += GET_MATRIX_DATA(iD, Y+13,X+2, inWS) * (-0.01029041f);
 result += GET_MATRIX_DATA(iD, Y+20,X+2, inWS) * (-0.05542325f);
 result += GET_MATRIX_DATA(iD, Y+5,X+3, inWS) * (-0.00352656f);
 result += GET_MATRIX_DATA(iD, Y+11,X+3, inWS) * (-0.01859207f);
 result += GET_MATRIX_DATA(iD, Y+17,X+3, inWS) * (-0.01258677f);
 result += GET_MATRIX_DATA(iD, Y+7,X+5, inWS) * (-0.01901069f);
 result += GET_MATRIX_DATA(iD, Y+15,X+5, inWS) * (-0.00509704f);
 result += GET_MATRIX_DATA(iD, Y+13,X+6, inWS) * (-0.03378659f);
 result += GET_MATRIX_DATA(iD, Y+15,X+7, inWS) * (-0.01940863f);
 result += GET_MATRIX_DATA(iD, Y+17,X+7, inWS) * (-0.06611530f);
 result += GET_MATRIX_DATA(iD, Y+23,X+7, inWS) * (-0.01601997f);
 result += GET_MATRIX_DATA(iD, Y+22,X+12, inWS) * (0.01000869f);
 result += GET_MATRIX_DATA(iD, Y+13,X+13, inWS) * (0.00606346f);
 result += GET_MATRIX_DATA(iD, Y+16,X+13, inWS) * (0.10822751f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (0.02986408f);
 result += GET_MATRIX_DATA(iD, Y+28,X+13, inWS) * (0.00420674f);
 result += GET_MATRIX_DATA(iD, Y+9,X+14, inWS) * (0.05223326f);
 result += GET_MATRIX_DATA(iD, Y+13,X+14, inWS) * (0.09382749f);
 result += GET_MATRIX_DATA(iD, Y+16,X+14, inWS) * (0.01203751f);
 result += GET_MATRIX_DATA(iD, Y+19,X+14, inWS) * (0.00095602f);
 result += GET_MATRIX_DATA(iD, Y+20,X+14, inWS) * (0.22484998f);
 result += GET_MATRIX_DATA(iD, Y+32,X+14, inWS) * (0.02620811f);
 result += GET_MATRIX_DATA(iD, Y+13,X+15, inWS) * (0.09220452f);
 result += GET_MATRIX_DATA(iD, Y+15,X+15, inWS) * (0.25726270f);
 result += GET_MATRIX_DATA(iD, Y+16,X+15, inWS) * (0.00773160f);
 result += GET_MATRIX_DATA(iD, Y+17,X+15, inWS) * (0.00190187f);
 result += GET_MATRIX_DATA(iD, Y+18,X+15, inWS) * (0.09411651f);
 result += GET_MATRIX_DATA(iD, Y+26,X+15, inWS) * (0.01069226f);
 result += GET_MATRIX_DATA(iD, Y+28,X+15, inWS) * (0.01205508f);
 result += GET_MATRIX_DATA(iD, Y+7,X+16, inWS) * (0.00927370f);
 result += GET_MATRIX_DATA(iD, Y+9,X+16, inWS) * (0.01897289f);
 result += GET_MATRIX_DATA(iD, Y+12,X+16, inWS) * (0.09900276f);
 result += GET_MATRIX_DATA(iD, Y+24,X+16, inWS) * (0.05753309f);
 result += GET_MATRIX_DATA(iD, Y+26,X+16, inWS) * (0.02660437f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (0.08687206f);
 result += GET_MATRIX_DATA(iD, Y+19,X+17, inWS) * (0.05852361f);
 result += GET_MATRIX_DATA(iD, Y+20,X+17, inWS) * (0.00928756f);
 result += GET_MATRIX_DATA(iD, Y+21,X+17, inWS) * (0.12712701f);
 result += GET_MATRIX_DATA(iD, Y+26,X+17, inWS) * (0.01580996f);
 result += GET_MATRIX_DATA(iD, Y+7,X+18, inWS) * (0.01119423f);
 result += GET_MATRIX_DATA(iD, Y+9,X+18, inWS) * (0.01922827f);
 result += GET_MATRIX_DATA(iD, Y+10,X+18, inWS) * (0.01299216f);
 result += GET_MATRIX_DATA(iD, Y+12,X+18, inWS) * (0.05554314f);
 result += GET_MATRIX_DATA(iD, Y+13,X+18, inWS) * (0.06155111f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (0.05328930f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (0.03046592f);
 result += GET_MATRIX_DATA(iD, Y+14,X+19, inWS) * (0.06858751f);
 result += GET_MATRIX_DATA(iD, Y+15,X+19, inWS) * (0.04059852f);
 result += GET_MATRIX_DATA(iD, Y+18,X+19, inWS) * (0.11484717f);
 result += GET_MATRIX_DATA(iD, Y+20,X+19, inWS) * (0.11193815f);
 result += GET_MATRIX_DATA(iD, Y+26,X+19, inWS) * (0.00363991f);
 result += GET_MATRIX_DATA(iD, Y+27,X+19, inWS) * (0.00226604f);
 result += GET_MATRIX_DATA(iD, Y+9,X+20, inWS) * (0.03730745f);
 result += GET_MATRIX_DATA(iD, Y+12,X+20, inWS) * (0.03180259f);
 result += GET_MATRIX_DATA(iD, Y+14,X+20, inWS) * (0.13313402f);
 result += GET_MATRIX_DATA(iD, Y+20,X+20, inWS) * (0.01382349f);
 result += GET_MATRIX_DATA(iD, Y+16,X+21, inWS) * (0.00674135f);
 result += GET_MATRIX_DATA(iD, Y+17,X+21, inWS) * (0.09814395f);
 result += GET_MATRIX_DATA(iD, Y+29,X+21, inWS) * (0.01020401f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.06179380f);
desc[2] = (result > (-990.64f));

 /***********************************************
                  Projection: 4
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+30,X+5, inWS) * (-0.00276578f);
 result += GET_MATRIX_DATA(iD, Y+24,X+7, inWS) * (-0.00509909f);
 result += GET_MATRIX_DATA(iD, Y+27,X+7, inWS) * (-0.00155486f);
 result += GET_MATRIX_DATA(iD, Y+2,X+8, inWS) * (0.03256770f);
 result += GET_MATRIX_DATA(iD, Y+28,X+8, inWS) * (-0.00523921f);
 result += GET_MATRIX_DATA(iD, Y+24,X+9, inWS) * (-0.01268650f);
 result += GET_MATRIX_DATA(iD, Y+28,X+9, inWS) * (-0.00794603f);
 result += GET_MATRIX_DATA(iD, Y+30,X+9, inWS) * (-0.02318462f);
 result += GET_MATRIX_DATA(iD, Y+5,X+10, inWS) * (0.01295614f);
 result += GET_MATRIX_DATA(iD, Y+7,X+10, inWS) * (0.00459118f);
 result += GET_MATRIX_DATA(iD, Y+3,X+12, inWS) * (0.01544978f);
 result += GET_MATRIX_DATA(iD, Y+30,X+12, inWS) * (-0.04581949f);
 result += GET_MATRIX_DATA(iD, Y+26,X+13, inWS) * (-0.01073703f);
 result += GET_MATRIX_DATA(iD, Y+3,X+14, inWS) * (0.07081354f);
 result += GET_MATRIX_DATA(iD, Y+12,X+15, inWS) * (-0.01658540f);
 result += GET_MATRIX_DATA(iD, Y+13,X+15, inWS) * (-0.11109255f);
 result += GET_MATRIX_DATA(iD, Y+19,X+15, inWS) * (0.07880416f);
 result += GET_MATRIX_DATA(iD, Y+21,X+15, inWS) * (0.05477763f);
 result += GET_MATRIX_DATA(iD, Y+26,X+15, inWS) * (-0.00279244f);
 result += GET_MATRIX_DATA(iD, Y+28,X+15, inWS) * (-0.02810341f);
 result += GET_MATRIX_DATA(iD, Y+19,X+16, inWS) * (0.14504866f);
 result += GET_MATRIX_DATA(iD, Y+28,X+16, inWS) * (-0.00127077f);
 result += GET_MATRIX_DATA(iD, Y+30,X+16, inWS) * (-0.07631253f);
 result += GET_MATRIX_DATA(iD, Y+12,X+17, inWS) * (-0.04582947f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (-0.02100798f);
 result += GET_MATRIX_DATA(iD, Y+14,X+17, inWS) * (-0.03319508f);
 result += GET_MATRIX_DATA(iD, Y+19,X+17, inWS) * (0.02976404f);
 result += GET_MATRIX_DATA(iD, Y+21,X+17, inWS) * (0.11927669f);
 result += GET_MATRIX_DATA(iD, Y+13,X+18, inWS) * (-0.11965395f);
 result += GET_MATRIX_DATA(iD, Y+21,X+18, inWS) * (0.05789107f);
 result += GET_MATRIX_DATA(iD, Y+2,X+19, inWS) * (0.00059213f);
 result += GET_MATRIX_DATA(iD, Y+5,X+19, inWS) * (0.05468416f);
 result += GET_MATRIX_DATA(iD, Y+14,X+19, inWS) * (-0.05991891f);
 result += GET_MATRIX_DATA(iD, Y+19,X+19, inWS) * (0.03199991f);
 result += GET_MATRIX_DATA(iD, Y+27,X+19, inWS) * (-0.00186910f);
 result += GET_MATRIX_DATA(iD, Y+2,X+20, inWS) * (0.02439495f);
 result += GET_MATRIX_DATA(iD, Y+19,X+20, inWS) * (0.01127270f);
 result += GET_MATRIX_DATA(iD, Y+27,X+20, inWS) * (-0.00104412f);
 result += GET_MATRIX_DATA(iD, Y+30,X+20, inWS) * (-0.03379836f);
 result += GET_MATRIX_DATA(iD, Y+31,X+20, inWS) * (-0.04166059f);
 result += GET_MATRIX_DATA(iD, Y+25,X+22, inWS) * (-0.00699648f);
 result += GET_MATRIX_DATA(iD, Y+26,X+22, inWS) * (-0.03562044f);
 result += GET_MATRIX_DATA(iD, Y+30,X+22, inWS) * (-0.01473083f);
 result += GET_MATRIX_DATA(iD, Y+1,X+23, inWS) * (0.00208679f);
 result += GET_MATRIX_DATA(iD, Y+2,X+23, inWS) * (0.00082987f);
 result += GET_MATRIX_DATA(iD, Y+3,X+23, inWS) * (0.00739773f);
 result += GET_MATRIX_DATA(iD, Y+8,X+23, inWS) * (0.00526382f);
 result += GET_MATRIX_DATA(iD, Y+27,X+24, inWS) * (-0.00507025f);
 result += GET_MATRIX_DATA(iD, Y+30,X+24, inWS) * (-0.04125883f);
 result += GET_MATRIX_DATA(iD, Y+8,X+25, inWS) * (0.00970654f);
 result += GET_MATRIX_DATA(iD, Y+3,X+26, inWS) * (0.02806777f);
 result += GET_MATRIX_DATA(iD, Y+27,X+26, inWS) * (-0.00048586f);
 result += GET_MATRIX_DATA(iD, Y+30,X+28, inWS) * (-0.03506485f);
 result += GET_MATRIX_DATA(iD, Y+3,X+30, inWS) * (0.01114135f);
 result += GET_MATRIX_DATA(iD, Y+7,X+30, inWS) * (0.00400613f);
 result += GET_MATRIX_DATA(iD, Y+24,X+30, inWS) * (-0.01660325f);
 result += GET_MATRIX_DATA(iD, Y+28,X+30, inWS) * (-0.02891329f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00166239f);
desc[3] = (result > (-86.59f));

 /***********************************************
                  Projection: 5
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+10,X+4, inWS) * (-0.00827370f);
 result += GET_MATRIX_DATA(iD, Y+11,X+6, inWS) * (-0.01049330f);
 result += GET_MATRIX_DATA(iD, Y+15,X+6, inWS) * (-0.00795969f);
 result += GET_MATRIX_DATA(iD, Y+21,X+6, inWS) * (-0.01254078f);
 result += GET_MATRIX_DATA(iD, Y+28,X+11, inWS) * (0.00238171f);
 result += GET_MATRIX_DATA(iD, Y+10,X+13, inWS) * (0.00575706f);
 result += GET_MATRIX_DATA(iD, Y+11,X+13, inWS) * (0.00718697f);
 result += GET_MATRIX_DATA(iD, Y+12,X+13, inWS) * (0.03048096f);
 result += GET_MATRIX_DATA(iD, Y+15,X+13, inWS) * (0.06541972f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (0.02344256f);
 result += GET_MATRIX_DATA(iD, Y+23,X+13, inWS) * (0.05601251f);
 result += GET_MATRIX_DATA(iD, Y+26,X+13, inWS) * (0.00256385f);
 result += GET_MATRIX_DATA(iD, Y+10,X+15, inWS) * (0.02933898f);
 result += GET_MATRIX_DATA(iD, Y+15,X+15, inWS) * (0.05913636f);
 result += GET_MATRIX_DATA(iD, Y+17,X+15, inWS) * (0.02424735f);
 result += GET_MATRIX_DATA(iD, Y+18,X+15, inWS) * (0.21178537f);
 result += GET_MATRIX_DATA(iD, Y+24,X+15, inWS) * (0.01866426f);
 result += GET_MATRIX_DATA(iD, Y+16,X+16, inWS) * (0.25327280f);
 result += GET_MATRIX_DATA(iD, Y+17,X+16, inWS) * (0.18648890f);
 result += GET_MATRIX_DATA(iD, Y+18,X+16, inWS) * (0.00883090f);
 result += GET_MATRIX_DATA(iD, Y+16,X+17, inWS) * (0.04847543f);
 result += GET_MATRIX_DATA(iD, Y+17,X+17, inWS) * (0.01142208f);
 result += GET_MATRIX_DATA(iD, Y+13,X+18, inWS) * (-0.09265488f);
 result += GET_MATRIX_DATA(iD, Y+14,X+18, inWS) * (-0.08613329f);
 result += GET_MATRIX_DATA(iD, Y+20,X+18, inWS) * (-0.11434023f);
 result += GET_MATRIX_DATA(iD, Y+21,X+18, inWS) * (-0.03224242f);
 result += GET_MATRIX_DATA(iD, Y+12,X+19, inWS) * (-0.01899471f);
 result += GET_MATRIX_DATA(iD, Y+13,X+19, inWS) * (-0.01302750f);
 result += GET_MATRIX_DATA(iD, Y+19,X+19, inWS) * (-0.07412544f);
 result += GET_MATRIX_DATA(iD, Y+20,X+19, inWS) * (-0.04676803f);
 result += GET_MATRIX_DATA(iD, Y+25,X+19, inWS) * (-0.03259860f);
 result += GET_MATRIX_DATA(iD, Y+26,X+19, inWS) * (-0.02426506f);
 result += GET_MATRIX_DATA(iD, Y+1,X+20, inWS) * (-0.00992396f);
 result += GET_MATRIX_DATA(iD, Y+4,X+20, inWS) * (-0.00803206f);
 result += GET_MATRIX_DATA(iD, Y+8,X+20, inWS) * (-0.00821592f);
 result += GET_MATRIX_DATA(iD, Y+9,X+20, inWS) * (-0.01688977f);
 result += GET_MATRIX_DATA(iD, Y+12,X+20, inWS) * (-0.04470068f);
 result += GET_MATRIX_DATA(iD, Y+15,X+20, inWS) * (-0.12468109f);
 result += GET_MATRIX_DATA(iD, Y+16,X+20, inWS) * (-0.11435803f);
 result += GET_MATRIX_DATA(iD, Y+17,X+20, inWS) * (-0.01753457f);
 result += GET_MATRIX_DATA(iD, Y+18,X+20, inWS) * (-0.14304509f);
 result += GET_MATRIX_DATA(iD, Y+19,X+20, inWS) * (-0.00887545f);
 result += GET_MATRIX_DATA(iD, Y+21,X+20, inWS) * (-0.01959245f);
 result += GET_MATRIX_DATA(iD, Y+7,X+21, inWS) * (-0.05865688f);
 result += GET_MATRIX_DATA(iD, Y+15,X+21, inWS) * (-0.02919327f);
 result += GET_MATRIX_DATA(iD, Y+16,X+21, inWS) * (-0.01460035f);
 result += GET_MATRIX_DATA(iD, Y+17,X+21, inWS) * (-0.07018450f);
 result += GET_MATRIX_DATA(iD, Y+23,X+21, inWS) * (-0.05294812f);
 result += GET_MATRIX_DATA(iD, Y+28,X+21, inWS) * (-0.00973854f);
 result += GET_MATRIX_DATA(iD, Y+16,X+27, inWS) * (0.00731143f);
 result += GET_MATRIX_DATA(iD, Y+12,X+30, inWS) * (0.00677607f);
 result += GET_MATRIX_DATA(iD, Y+18,X+30, inWS) * (0.00231903f);
 result += GET_MATRIX_DATA(iD, Y+22,X+30, inWS) * (0.01616258f);
 result += GET_MATRIX_DATA(iD, Y+24,X+30, inWS) * (0.01279543f);
 result += GET_MATRIX_DATA(iD, Y+1,X+31, inWS) * (0.00685273f);
 result += GET_MATRIX_DATA(iD, Y+15,X+31, inWS) * (0.00152107f);
 result += GET_MATRIX_DATA(iD, Y+16,X+31, inWS) * (0.03480435f);
 result += GET_MATRIX_DATA(iD, Y+10,X+32, inWS) * (0.02764804f);
 result += GET_MATRIX_DATA(iD, Y+26,X+32, inWS) * (0.00811322f);
 result += GET_MATRIX_DATA(iD, Y+30,X+32, inWS) * (0.02226667f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00035622f);
desc[4] = (result > (-872.15f));

 /***********************************************
                  Projection: 6
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+21,X+3, inWS) * (-0.01658711f);
 result += GET_MATRIX_DATA(iD, Y+16,X+5, inWS) * (-0.00276673f);
 result += GET_MATRIX_DATA(iD, Y+18,X+5, inWS) * (-0.02389054f);
 result += GET_MATRIX_DATA(iD, Y+19,X+5, inWS) * (-0.00744475f);
 result += GET_MATRIX_DATA(iD, Y+21,X+5, inWS) * (-0.00608324f);
 result += GET_MATRIX_DATA(iD, Y+23,X+5, inWS) * (-0.00302022f);
 result += GET_MATRIX_DATA(iD, Y+17,X+7, inWS) * (-0.02451381f);
 result += GET_MATRIX_DATA(iD, Y+20,X+7, inWS) * (-0.00284094f);
 result += GET_MATRIX_DATA(iD, Y+23,X+7, inWS) * (-0.00877955f);
 result += GET_MATRIX_DATA(iD, Y+3,X+8, inWS) * (0.01159428f);
 result += GET_MATRIX_DATA(iD, Y+20,X+8, inWS) * (-0.00642500f);
 result += GET_MATRIX_DATA(iD, Y+18,X+9, inWS) * (-0.03826959f);
 result += GET_MATRIX_DATA(iD, Y+20,X+9, inWS) * (-0.00043301f);
 result += GET_MATRIX_DATA(iD, Y+3,X+10, inWS) * (0.02962400f);
 result += GET_MATRIX_DATA(iD, Y+16,X+10, inWS) * (-0.05271816f);
 result += GET_MATRIX_DATA(iD, Y+18,X+10, inWS) * (-0.02876937f);
 result += GET_MATRIX_DATA(iD, Y+6,X+11, inWS) * (0.00783842f);
 result += GET_MATRIX_DATA(iD, Y+18,X+11, inWS) * (-0.02962044f);
 result += GET_MATRIX_DATA(iD, Y+3,X+12, inWS) * (0.04556902f);
 result += GET_MATRIX_DATA(iD, Y+8,X+13, inWS) * (0.07312218f);
 result += GET_MATRIX_DATA(iD, Y+9,X+13, inWS) * (0.00977408f);
 result += GET_MATRIX_DATA(iD, Y+30,X+13, inWS) * (0.00952143f);
 result += GET_MATRIX_DATA(iD, Y+32,X+13, inWS) * (0.01432877f);
 result += GET_MATRIX_DATA(iD, Y+3,X+14, inWS) * (0.02169598f);
 result += GET_MATRIX_DATA(iD, Y+26,X+14, inWS) * (0.00394955f);
 result += GET_MATRIX_DATA(iD, Y+9,X+15, inWS) * (0.02478136f);
 result += GET_MATRIX_DATA(iD, Y+32,X+15, inWS) * (0.00526530f);
 result += GET_MATRIX_DATA(iD, Y+3,X+16, inWS) * (0.06306729f);
 result += GET_MATRIX_DATA(iD, Y+7,X+16, inWS) * (0.02909606f);
 result += GET_MATRIX_DATA(iD, Y+9,X+16, inWS) * (0.00676241f);
 result += GET_MATRIX_DATA(iD, Y+16,X+16, inWS) * (-0.01955875f);
 result += GET_MATRIX_DATA(iD, Y+18,X+16, inWS) * (-0.02317537f);
 result += GET_MATRIX_DATA(iD, Y+21,X+16, inWS) * (0.00443655f);
 result += GET_MATRIX_DATA(iD, Y+26,X+16, inWS) * (0.00658951f);
 result += GET_MATRIX_DATA(iD, Y+27,X+16, inWS) * (0.01805335f);
 result += GET_MATRIX_DATA(iD, Y+28,X+16, inWS) * (0.03136585f);
 result += GET_MATRIX_DATA(iD, Y+21,X+17, inWS) * (0.04085900f);
 result += GET_MATRIX_DATA(iD, Y+26,X+17, inWS) * (0.00657839f);
 result += GET_MATRIX_DATA(iD, Y+32,X+17, inWS) * (0.03799442f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (-0.02451281f);
 result += GET_MATRIX_DATA(iD, Y+16,X+18, inWS) * (-0.15428609f);
 result += GET_MATRIX_DATA(iD, Y+17,X+18, inWS) * (-0.16036075f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (-0.04285502f);
 result += GET_MATRIX_DATA(iD, Y+28,X+18, inWS) * (0.01086563f);
 result += GET_MATRIX_DATA(iD, Y+30,X+19, inWS) * (0.00275750f);
 result += GET_MATRIX_DATA(iD, Y+15,X+20, inWS) * (-0.00966775f);
 result += GET_MATRIX_DATA(iD, Y+28,X+22, inWS) * (0.02358243f);
 result += GET_MATRIX_DATA(iD, Y+30,X+22, inWS) * (0.00144983f);
 result += GET_MATRIX_DATA(iD, Y+11,X+23, inWS) * (-0.02311022f);
 result += GET_MATRIX_DATA(iD, Y+12,X+23, inWS) * (-0.00286465f);
 result += GET_MATRIX_DATA(iD, Y+17,X+23, inWS) * (-0.03829666f);
 result += GET_MATRIX_DATA(iD, Y+19,X+23, inWS) * (-0.07198641f);
 result += GET_MATRIX_DATA(iD, Y+15,X+24, inWS) * (-0.06753573f);
 result += GET_MATRIX_DATA(iD, Y+16,X+24, inWS) * (-0.01147935f);
 result += GET_MATRIX_DATA(iD, Y+14,X+25, inWS) * (-0.03870888f);
 result += GET_MATRIX_DATA(iD, Y+14,X+27, inWS) * (-0.00568482f);
 result += GET_MATRIX_DATA(iD, Y+17,X+27, inWS) * (-0.00748945f);
 result += GET_MATRIX_DATA(iD, Y+15,X+29, inWS) * (-0.00423372f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.01170483f);
desc[5] = (result > (20.74f));

 /***********************************************
                  Projection: 7
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+5,X+5, inWS) * (-0.00877725f);
 result += GET_MATRIX_DATA(iD, Y+23,X+5, inWS) * (0.00653078f);
 result += GET_MATRIX_DATA(iD, Y+28,X+6, inWS) * (0.00372219f);
 result += GET_MATRIX_DATA(iD, Y+27,X+7, inWS) * (0.00656050f);
 result += GET_MATRIX_DATA(iD, Y+6,X+8, inWS) * (-0.00193168f);
 result += GET_MATRIX_DATA(iD, Y+3,X+9, inWS) * (-0.01808935f);
 result += GET_MATRIX_DATA(iD, Y+22,X+9, inWS) * (0.01198614f);
 result += GET_MATRIX_DATA(iD, Y+23,X+9, inWS) * (0.00289672f);
 result += GET_MATRIX_DATA(iD, Y+24,X+9, inWS) * (0.00075324f);
 result += GET_MATRIX_DATA(iD, Y+3,X+10, inWS) * (-0.00583130f);
 result += GET_MATRIX_DATA(iD, Y+24,X+10, inWS) * (0.01649486f);
 result += GET_MATRIX_DATA(iD, Y+25,X+10, inWS) * (0.00474115f);
 result += GET_MATRIX_DATA(iD, Y+22,X+11, inWS) * (0.03094702f);
 result += GET_MATRIX_DATA(iD, Y+6,X+12, inWS) * (-0.02992440f);
 result += GET_MATRIX_DATA(iD, Y+22,X+12, inWS) * (0.00388548f);
 result += GET_MATRIX_DATA(iD, Y+24,X+12, inWS) * (0.00193172f);
 result += GET_MATRIX_DATA(iD, Y+3,X+13, inWS) * (-0.01032963f);
 result += GET_MATRIX_DATA(iD, Y+7,X+13, inWS) * (-0.00147341f);
 result += GET_MATRIX_DATA(iD, Y+8,X+13, inWS) * (-0.02215016f);
 result += GET_MATRIX_DATA(iD, Y+16,X+13, inWS) * (0.04361915f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (0.09738749f);
 result += GET_MATRIX_DATA(iD, Y+18,X+13, inWS) * (0.07046285f);
 result += GET_MATRIX_DATA(iD, Y+20,X+14, inWS) * (0.05924122f);
 result += GET_MATRIX_DATA(iD, Y+13,X+15, inWS) * (0.03087605f);
 result += GET_MATRIX_DATA(iD, Y+20,X+15, inWS) * (0.04335502f);
 result += GET_MATRIX_DATA(iD, Y+9,X+16, inWS) * (-0.01098926f);
 result += GET_MATRIX_DATA(iD, Y+11,X+16, inWS) * (-0.01678953f);
 result += GET_MATRIX_DATA(iD, Y+18,X+16, inWS) * (-0.05074493f);
 result += GET_MATRIX_DATA(iD, Y+20,X+16, inWS) * (0.02266847f);
 result += GET_MATRIX_DATA(iD, Y+21,X+16, inWS) * (0.07545620f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (0.08657491f);
 result += GET_MATRIX_DATA(iD, Y+16,X+17, inWS) * (-0.21861581f);
 result += GET_MATRIX_DATA(iD, Y+20,X+17, inWS) * (0.00171185f);
 result += GET_MATRIX_DATA(iD, Y+23,X+17, inWS) * (-0.01141346f);
 result += GET_MATRIX_DATA(iD, Y+16,X+18, inWS) * (-0.05915727f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (-0.14771125f);
 result += GET_MATRIX_DATA(iD, Y+23,X+19, inWS) * (-0.08037894f);
 result += GET_MATRIX_DATA(iD, Y+28,X+19, inWS) * (-0.00386788f);
 result += GET_MATRIX_DATA(iD, Y+28,X+20, inWS) * (-0.01229163f);
 result += GET_MATRIX_DATA(iD, Y+16,X+21, inWS) * (0.04386875f);
 result += GET_MATRIX_DATA(iD, Y+22,X+21, inWS) * (-0.04098269f);
 result += GET_MATRIX_DATA(iD, Y+25,X+21, inWS) * (-0.00339696f);
 result += GET_MATRIX_DATA(iD, Y+28,X+22, inWS) * (-0.00200411f);
 result += GET_MATRIX_DATA(iD, Y+11,X+23, inWS) * (0.01524189f);
 result += GET_MATRIX_DATA(iD, Y+24,X+23, inWS) * (-0.00325058f);
 result += GET_MATRIX_DATA(iD, Y+26,X+23, inWS) * (-0.02001108f);
 result += GET_MATRIX_DATA(iD, Y+7,X+24, inWS) * (0.00239294f);
 result += GET_MATRIX_DATA(iD, Y+22,X+24, inWS) * (-0.00015200f);
 result += GET_MATRIX_DATA(iD, Y+22,X+25, inWS) * (-0.00486056f);
 result += GET_MATRIX_DATA(iD, Y+25,X+25, inWS) * (-0.00470970f);
 result += GET_MATRIX_DATA(iD, Y+9,X+26, inWS) * (0.02634842f);
 result += GET_MATRIX_DATA(iD, Y+3,X+27, inWS) * (0.00638850f);
 result += GET_MATRIX_DATA(iD, Y+15,X+27, inWS) * (0.00705495f);
 result += GET_MATRIX_DATA(iD, Y+3,X+28, inWS) * (0.00401791f);
 result += GET_MATRIX_DATA(iD, Y+11,X+28, inWS) * (0.01833133f);
 result += GET_MATRIX_DATA(iD, Y+7,X+30, inWS) * (0.02718795f);
 result += GET_MATRIX_DATA(iD, Y+13,X+30, inWS) * (0.00992100f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00071288f);
desc[6] = (result > (-238.68f));

 /***********************************************
                  Projection: 8
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+31,X+2, inWS) * (0.00710296f);
 result += GET_MATRIX_DATA(iD, Y+8,X+3, inWS) * (-0.00038033f);
 result += GET_MATRIX_DATA(iD, Y+9,X+3, inWS) * (-0.00858161f);
 result += GET_MATRIX_DATA(iD, Y+27,X+3, inWS) * (0.01400175f);
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (0.01417541f);
 result += GET_MATRIX_DATA(iD, Y+8,X+5, inWS) * (-0.02193843f);
 result += GET_MATRIX_DATA(iD, Y+12,X+7, inWS) * (-0.00649011f);
 result += GET_MATRIX_DATA(iD, Y+14,X+7, inWS) * (-0.02293599f);
 result += GET_MATRIX_DATA(iD, Y+15,X+7, inWS) * (-0.02819053f);
 result += GET_MATRIX_DATA(iD, Y+25,X+7, inWS) * (0.00156716f);
 result += GET_MATRIX_DATA(iD, Y+28,X+7, inWS) * (0.02663511f);
 result += GET_MATRIX_DATA(iD, Y+9,X+8, inWS) * (-0.00412611f);
 result += GET_MATRIX_DATA(iD, Y+12,X+8, inWS) * (-0.00360457f);
 result += GET_MATRIX_DATA(iD, Y+30,X+8, inWS) * (0.05104896f);
 result += GET_MATRIX_DATA(iD, Y+32,X+8, inWS) * (0.00428549f);
 result += GET_MATRIX_DATA(iD, Y+11,X+9, inWS) * (-0.00610062f);
 result += GET_MATRIX_DATA(iD, Y+11,X+10, inWS) * (-0.02575856f);
 result += GET_MATRIX_DATA(iD, Y+26,X+10, inWS) * (0.06136231f);
 result += GET_MATRIX_DATA(iD, Y+12,X+11, inWS) * (-0.01269103f);
 result += GET_MATRIX_DATA(iD, Y+30,X+11, inWS) * (0.00075851f);
 result += GET_MATRIX_DATA(iD, Y+32,X+11, inWS) * (0.05917990f);
 result += GET_MATRIX_DATA(iD, Y+12,X+12, inWS) * (-0.00643373f);
 result += GET_MATRIX_DATA(iD, Y+16,X+12, inWS) * (-0.07766062f);
 result += GET_MATRIX_DATA(iD, Y+18,X+12, inWS) * (-0.02298025f);
 result += GET_MATRIX_DATA(iD, Y+26,X+12, inWS) * (0.00737998f);
 result += GET_MATRIX_DATA(iD, Y+28,X+12, inWS) * (0.02793385f);
 result += GET_MATRIX_DATA(iD, Y+14,X+13, inWS) * (-0.01430495f);
 result += GET_MATRIX_DATA(iD, Y+16,X+13, inWS) * (-0.10365330f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (-0.01304104f);
 result += GET_MATRIX_DATA(iD, Y+24,X+13, inWS) * (0.04029112f);
 result += GET_MATRIX_DATA(iD, Y+30,X+13, inWS) * (0.00621063f);
 result += GET_MATRIX_DATA(iD, Y+14,X+14, inWS) * (-0.06101499f);
 result += GET_MATRIX_DATA(iD, Y+28,X+14, inWS) * (0.00762311f);
 result += GET_MATRIX_DATA(iD, Y+30,X+14, inWS) * (0.03547384f);
 result += GET_MATRIX_DATA(iD, Y+26,X+15, inWS) * (0.00440744f);
 result += GET_MATRIX_DATA(iD, Y+27,X+15, inWS) * (0.00518967f);
 result += GET_MATRIX_DATA(iD, Y+28,X+15, inWS) * (0.03741135f);
 result += GET_MATRIX_DATA(iD, Y+5,X+16, inWS) * (0.01502481f);
 result += GET_MATRIX_DATA(iD, Y+17,X+17, inWS) * (0.01497752f);
 result += GET_MATRIX_DATA(iD, Y+6,X+18, inWS) * (0.00095046f);
 result += GET_MATRIX_DATA(iD, Y+10,X+19, inWS) * (0.09130017f);
 result += GET_MATRIX_DATA(iD, Y+21,X+19, inWS) * (-0.02256498f);
 result += GET_MATRIX_DATA(iD, Y+3,X+20, inWS) * (0.00926131f);
 result += GET_MATRIX_DATA(iD, Y+5,X+20, inWS) * (0.00513772f);
 result += GET_MATRIX_DATA(iD, Y+6,X+20, inWS) * (0.00201725f);
 result += GET_MATRIX_DATA(iD, Y+7,X+20, inWS) * (0.03603126f);
 result += GET_MATRIX_DATA(iD, Y+9,X+20, inWS) * (0.00595592f);
 result += GET_MATRIX_DATA(iD, Y+3,X+21, inWS) * (0.01604307f);
 result += GET_MATRIX_DATA(iD, Y+19,X+21, inWS) * (-0.00720328f);
 result += GET_MATRIX_DATA(iD, Y+21,X+21, inWS) * (-0.01605866f);
 result += GET_MATRIX_DATA(iD, Y+2,X+23, inWS) * (0.00172277f);
 result += GET_MATRIX_DATA(iD, Y+10,X+23, inWS) * (0.02128276f);
 result += GET_MATRIX_DATA(iD, Y+4,X+24, inWS) * (0.00243290f);
 result += GET_MATRIX_DATA(iD, Y+7,X+24, inWS) * (0.00566436f);
 result += GET_MATRIX_DATA(iD, Y+8,X+24, inWS) * (0.01430793f);
 result += GET_MATRIX_DATA(iD, Y+22,X+25, inWS) * (-0.04923641f);
 result += GET_MATRIX_DATA(iD, Y+3,X+26, inWS) * (0.00086680f);
 result += GET_MATRIX_DATA(iD, Y+6,X+26, inWS) * (0.00601368f);
 result += GET_MATRIX_DATA(iD, Y+19,X+28, inWS) * (-0.00529531f);
 result += GET_MATRIX_DATA(iD, Y+25,X+28, inWS) * (-0.00440428f);
 result += GET_MATRIX_DATA(iD, Y+19,X+30, inWS) * (-0.01508551f);
 result += GET_MATRIX_DATA(iD, Y+21,X+30, inWS) * (-0.01029177f);
 result += GET_MATRIX_DATA(iD, Y+26,X+30, inWS) * (-0.01820490f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00076139f);
desc[7] = (result > (313.21f));

 /***********************************************
                  Projection: 9
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+3,X+2, inWS) * (0.00667533f);
 result += GET_MATRIX_DATA(iD, Y+5,X+2, inWS) * (0.00273720f);
 result += GET_MATRIX_DATA(iD, Y+7,X+3, inWS) * (0.00805491f);
 result += GET_MATRIX_DATA(iD, Y+24,X+3, inWS) * (-0.00074078f);
 result += GET_MATRIX_DATA(iD, Y+3,X+4, inWS) * (0.00126785f);
 result += GET_MATRIX_DATA(iD, Y+3,X+5, inWS) * (0.01244523f);
 result += GET_MATRIX_DATA(iD, Y+3,X+8, inWS) * (0.01397271f);
 result += GET_MATRIX_DATA(iD, Y+27,X+13, inWS) * (0.01807445f);
 result += GET_MATRIX_DATA(iD, Y+28,X+13, inWS) * (0.01537256f);
 result += GET_MATRIX_DATA(iD, Y+9,X+14, inWS) * (-0.00247823f);
 result += GET_MATRIX_DATA(iD, Y+18,X+14, inWS) * (-0.00857017f);
 result += GET_MATRIX_DATA(iD, Y+19,X+14, inWS) * (-0.00151464f);
 result += GET_MATRIX_DATA(iD, Y+20,X+14, inWS) * (-0.04887071f);
 result += GET_MATRIX_DATA(iD, Y+7,X+15, inWS) * (-0.03238066f);
 result += GET_MATRIX_DATA(iD, Y+18,X+15, inWS) * (-0.11934485f);
 result += GET_MATRIX_DATA(iD, Y+25,X+15, inWS) * (0.05208697f);
 result += GET_MATRIX_DATA(iD, Y+26,X+15, inWS) * (0.02997628f);
 result += GET_MATRIX_DATA(iD, Y+30,X+15, inWS) * (0.03592747f);
 result += GET_MATRIX_DATA(iD, Y+7,X+16, inWS) * (-0.01278566f);
 result += GET_MATRIX_DATA(iD, Y+9,X+16, inWS) * (-0.00153831f);
 result += GET_MATRIX_DATA(iD, Y+15,X+16, inWS) * (0.03971977f);
 result += GET_MATRIX_DATA(iD, Y+16,X+16, inWS) * (0.02477548f);
 result += GET_MATRIX_DATA(iD, Y+25,X+16, inWS) * (0.00170352f);
 result += GET_MATRIX_DATA(iD, Y+27,X+16, inWS) * (0.00924572f);
 result += GET_MATRIX_DATA(iD, Y+28,X+16, inWS) * (0.00852288f);
 result += GET_MATRIX_DATA(iD, Y+5,X+17, inWS) * (-0.00565190f);
 result += GET_MATRIX_DATA(iD, Y+7,X+17, inWS) * (-0.00393094f);
 result += GET_MATRIX_DATA(iD, Y+9,X+17, inWS) * (-0.02625117f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (0.04119192f);
 result += GET_MATRIX_DATA(iD, Y+15,X+17, inWS) * (0.02589111f);
 result += GET_MATRIX_DATA(iD, Y+16,X+17, inWS) * (0.00521079f);
 result += GET_MATRIX_DATA(iD, Y+24,X+17, inWS) * (0.01606134f);
 result += GET_MATRIX_DATA(iD, Y+25,X+17, inWS) * (0.01348851f);
 result += GET_MATRIX_DATA(iD, Y+3,X+18, inWS) * (-0.01143835f);
 result += GET_MATRIX_DATA(iD, Y+9,X+18, inWS) * (-0.04221932f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (0.11634136f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (-0.07419462f);
 result += GET_MATRIX_DATA(iD, Y+20,X+18, inWS) * (-0.04586211f);
 result += GET_MATRIX_DATA(iD, Y+3,X+19, inWS) * (-0.01816959f);
 result += GET_MATRIX_DATA(iD, Y+15,X+19, inWS) * (0.02073670f);
 result += GET_MATRIX_DATA(iD, Y+16,X+19, inWS) * (0.01003544f);
 result += GET_MATRIX_DATA(iD, Y+6,X+20, inWS) * (-0.02357733f);
 result += GET_MATRIX_DATA(iD, Y+7,X+20, inWS) * (-0.00720479f);
 result += GET_MATRIX_DATA(iD, Y+8,X+20, inWS) * (-0.05506688f);
 result += GET_MATRIX_DATA(iD, Y+10,X+20, inWS) * (-0.00255680f);
 result += GET_MATRIX_DATA(iD, Y+4,X+21, inWS) * (-0.00312104f);
 result += GET_MATRIX_DATA(iD, Y+15,X+22, inWS) * (0.04618904f);
 result += GET_MATRIX_DATA(iD, Y+26,X+26, inWS) * (-0.00044853f);
 result += GET_MATRIX_DATA(iD, Y+28,X+26, inWS) * (-0.02699554f);
 result += GET_MATRIX_DATA(iD, Y+12,X+28, inWS) * (0.00835012f);
 result += GET_MATRIX_DATA(iD, Y+15,X+28, inWS) * (0.00092959f);
 result += GET_MATRIX_DATA(iD, Y+26,X+28, inWS) * (-0.00073829f);
 result += GET_MATRIX_DATA(iD, Y+30,X+28, inWS) * (-0.00432528f);
 result += GET_MATRIX_DATA(iD, Y+32,X+28, inWS) * (-0.00352425f);
 result += GET_MATRIX_DATA(iD, Y+7,X+30, inWS) * (0.00358385f);
 result += GET_MATRIX_DATA(iD, Y+10,X+30, inWS) * (0.01269228f);
 result += GET_MATRIX_DATA(iD, Y+12,X+30, inWS) * (0.00053766f);
 result += GET_MATRIX_DATA(iD, Y+13,X+30, inWS) * (0.00620775f);
 result += GET_MATRIX_DATA(iD, Y+28,X+30, inWS) * (-0.01310295f);
 result += GET_MATRIX_DATA(iD, Y+30,X+30, inWS) * (-0.01727429f);
 result += GET_MATRIX_DATA(iD, Y+9,X+32, inWS) * (0.01094037f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00033082f);
desc[8] = (result > (54.09f));

 /***********************************************
                  Projection: 10
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+16,X+1, inWS) * (-0.00392271f);
 result += GET_MATRIX_DATA(iD, Y+17,X+1, inWS) * (-0.00252199f);
 result += GET_MATRIX_DATA(iD, Y+19,X+1, inWS) * (-0.02593446f);
 result += GET_MATRIX_DATA(iD, Y+21,X+1, inWS) * (-0.00928202f);
 result += GET_MATRIX_DATA(iD, Y+24,X+1, inWS) * (-0.04105111f);
 result += GET_MATRIX_DATA(iD, Y+30,X+2, inWS) * (-0.01014201f);
 result += GET_MATRIX_DATA(iD, Y+28,X+3, inWS) * (-0.00285748f);
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (-0.01157019f);
 result += GET_MATRIX_DATA(iD, Y+17,X+8, inWS) * (0.01480492f);
 result += GET_MATRIX_DATA(iD, Y+12,X+9, inWS) * (0.00906150f);
 result += GET_MATRIX_DATA(iD, Y+14,X+9, inWS) * (0.01682713f);
 result += GET_MATRIX_DATA(iD, Y+15,X+9, inWS) * (0.00490009f);
 result += GET_MATRIX_DATA(iD, Y+18,X+9, inWS) * (0.00916208f);
 result += GET_MATRIX_DATA(iD, Y+21,X+9, inWS) * (0.02706854f);
 result += GET_MATRIX_DATA(iD, Y+12,X+10, inWS) * (0.03153018f);
 result += GET_MATRIX_DATA(iD, Y+16,X+10, inWS) * (0.06150391f);
 result += GET_MATRIX_DATA(iD, Y+18,X+10, inWS) * (0.01150285f);
 result += GET_MATRIX_DATA(iD, Y+19,X+10, inWS) * (0.02710561f);
 result += GET_MATRIX_DATA(iD, Y+19,X+11, inWS) * (0.00106336f);
 result += GET_MATRIX_DATA(iD, Y+21,X+11, inWS) * (0.00238252f);
 result += GET_MATRIX_DATA(iD, Y+23,X+12, inWS) * (0.00151604f);
 result += GET_MATRIX_DATA(iD, Y+23,X+13, inWS) * (0.00326665f);
 result += GET_MATRIX_DATA(iD, Y+13,X+16, inWS) * (-0.02058890f);
 result += GET_MATRIX_DATA(iD, Y+26,X+16, inWS) * (0.00695539f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (-0.08646957f);
 result += GET_MATRIX_DATA(iD, Y+21,X+17, inWS) * (-0.01901069f);
 result += GET_MATRIX_DATA(iD, Y+26,X+17, inWS) * (0.00230374f);
 result += GET_MATRIX_DATA(iD, Y+14,X+18, inWS) * (-0.02939020f);
 result += GET_MATRIX_DATA(iD, Y+19,X+18, inWS) * (-0.00918635f);
 result += GET_MATRIX_DATA(iD, Y+20,X+18, inWS) * (-0.15851305f);
 result += GET_MATRIX_DATA(iD, Y+17,X+19, inWS) * (-0.12661672f);
 result += GET_MATRIX_DATA(iD, Y+26,X+19, inWS) * (0.02584661f);
 result += GET_MATRIX_DATA(iD, Y+26,X+20, inWS) * (0.00845192f);
 result += GET_MATRIX_DATA(iD, Y+27,X+20, inWS) * (0.00499106f);
 result += GET_MATRIX_DATA(iD, Y+28,X+20, inWS) * (0.01021303f);
 result += GET_MATRIX_DATA(iD, Y+30,X+20, inWS) * (0.00179045f);
 result += GET_MATRIX_DATA(iD, Y+28,X+22, inWS) * (0.00568150f);
 result += GET_MATRIX_DATA(iD, Y+10,X+23, inWS) * (0.02108218f);
 result += GET_MATRIX_DATA(iD, Y+12,X+23, inWS) * (0.03004715f);
 result += GET_MATRIX_DATA(iD, Y+13,X+23, inWS) * (0.00401034f);
 result += GET_MATRIX_DATA(iD, Y+15,X+23, inWS) * (0.00392888f);
 result += GET_MATRIX_DATA(iD, Y+16,X+23, inWS) * (0.03083892f);
 result += GET_MATRIX_DATA(iD, Y+17,X+23, inWS) * (0.02829298f);
 result += GET_MATRIX_DATA(iD, Y+22,X+23, inWS) * (0.01572127f);
 result += GET_MATRIX_DATA(iD, Y+24,X+23, inWS) * (0.02950052f);
 result += GET_MATRIX_DATA(iD, Y+15,X+24, inWS) * (0.01200276f);
 result += GET_MATRIX_DATA(iD, Y+20,X+24, inWS) * (0.01538226f);
 result += GET_MATRIX_DATA(iD, Y+11,X+30, inWS) * (-0.00600187f);
 result += GET_MATRIX_DATA(iD, Y+25,X+30, inWS) * (-0.00757189f);
 result += GET_MATRIX_DATA(iD, Y+14,X+31, inWS) * (-0.00856873f);
 result += GET_MATRIX_DATA(iD, Y+19,X+31, inWS) * (-0.00985359f);
 result += GET_MATRIX_DATA(iD, Y+22,X+31, inWS) * (-0.00890602f);
 result += GET_MATRIX_DATA(iD, Y+31,X+31, inWS) * (-0.00184831f);
 result += GET_MATRIX_DATA(iD, Y+25,X+32, inWS) * (-0.00699633f);
 result += GET_MATRIX_DATA(iD, Y+28,X+32, inWS) * (-0.01379149f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00264316f);
desc[9] = (result > (-677.07f));

 /***********************************************
                  Projection: 11
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+3,X+1, inWS) * (0.01291955f);
 result += GET_MATRIX_DATA(iD, Y+7,X+1, inWS) * (0.02254501f);
 result += GET_MATRIX_DATA(iD, Y+9,X+1, inWS) * (0.00365433f);
 result += GET_MATRIX_DATA(iD, Y+10,X+1, inWS) * (0.00121809f);
 result += GET_MATRIX_DATA(iD, Y+26,X+1, inWS) * (0.01128622f);
 result += GET_MATRIX_DATA(iD, Y+30,X+1, inWS) * (0.02056020f);
 result += GET_MATRIX_DATA(iD, Y+11,X+2, inWS) * (0.00039861f);
 result += GET_MATRIX_DATA(iD, Y+1,X+3, inWS) * (0.01396029f);
 result += GET_MATRIX_DATA(iD, Y+3,X+3, inWS) * (0.03527467f);
 result += GET_MATRIX_DATA(iD, Y+7,X+3, inWS) * (0.00576222f);
 result += GET_MATRIX_DATA(iD, Y+11,X+3, inWS) * (0.00075304f);
 result += GET_MATRIX_DATA(iD, Y+19,X+7, inWS) * (-0.00135597f);
 result += GET_MATRIX_DATA(iD, Y+20,X+7, inWS) * (-0.00478704f);
 result += GET_MATRIX_DATA(iD, Y+11,X+8, inWS) * (-0.01500391f);
 result += GET_MATRIX_DATA(iD, Y+17,X+8, inWS) * (-0.00501920f);
 result += GET_MATRIX_DATA(iD, Y+18,X+8, inWS) * (-0.02500944f);
 result += GET_MATRIX_DATA(iD, Y+13,X+9, inWS) * (-0.03444128f);
 result += GET_MATRIX_DATA(iD, Y+16,X+9, inWS) * (-0.01544272f);
 result += GET_MATRIX_DATA(iD, Y+21,X+9, inWS) * (-0.04917971f);
 result += GET_MATRIX_DATA(iD, Y+23,X+9, inWS) * (-0.01148696f);
 result += GET_MATRIX_DATA(iD, Y+8,X+10, inWS) * (-0.00292059f);
 result += GET_MATRIX_DATA(iD, Y+10,X+10, inWS) * (-0.00552182f);
 result += GET_MATRIX_DATA(iD, Y+12,X+10, inWS) * (-0.02286655f);
 result += GET_MATRIX_DATA(iD, Y+13,X+10, inWS) * (-0.00944166f);
 result += GET_MATRIX_DATA(iD, Y+10,X+11, inWS) * (-0.04286512f);
 result += GET_MATRIX_DATA(iD, Y+14,X+11, inWS) * (-0.01251500f);
 result += GET_MATRIX_DATA(iD, Y+24,X+11, inWS) * (-0.01407219f);
 result += GET_MATRIX_DATA(iD, Y+26,X+11, inWS) * (-0.03601132f);
 result += GET_MATRIX_DATA(iD, Y+6,X+12, inWS) * (-0.00082124f);
 result += GET_MATRIX_DATA(iD, Y+6,X+13, inWS) * (-0.01906440f);
 result += GET_MATRIX_DATA(iD, Y+8,X+13, inWS) * (-0.05275741f);
 result += GET_MATRIX_DATA(iD, Y+28,X+13, inWS) * (-0.00580133f);
 result += GET_MATRIX_DATA(iD, Y+30,X+13, inWS) * (-0.01234668f);
 result += GET_MATRIX_DATA(iD, Y+16,X+14, inWS) * (0.03146491f);
 result += GET_MATRIX_DATA(iD, Y+17,X+14, inWS) * (0.06473700f);
 result += GET_MATRIX_DATA(iD, Y+20,X+14, inWS) * (0.00934180f);
 result += GET_MATRIX_DATA(iD, Y+8,X+15, inWS) * (-0.00716936f);
 result += GET_MATRIX_DATA(iD, Y+16,X+15, inWS) * (0.01727108f);
 result += GET_MATRIX_DATA(iD, Y+17,X+15, inWS) * (0.01944043f);
 result += GET_MATRIX_DATA(iD, Y+15,X+16, inWS) * (0.03269329f);
 result += GET_MATRIX_DATA(iD, Y+16,X+16, inWS) * (0.03631740f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (-0.07424622f);
 result += GET_MATRIX_DATA(iD, Y+23,X+18, inWS) * (0.00376105f);
 result += GET_MATRIX_DATA(iD, Y+11,X+19, inWS) * (0.00718095f);
 result += GET_MATRIX_DATA(iD, Y+17,X+19, inWS) * (-0.07476400f);
 result += GET_MATRIX_DATA(iD, Y+18,X+19, inWS) * (-0.02407675f);
 result += GET_MATRIX_DATA(iD, Y+19,X+19, inWS) * (-0.02304222f);
 result += GET_MATRIX_DATA(iD, Y+24,X+19, inWS) * (0.00957392f);
 result += GET_MATRIX_DATA(iD, Y+11,X+20, inWS) * (0.00821583f);
 result += GET_MATRIX_DATA(iD, Y+17,X+20, inWS) * (-0.00473969f);
 result += GET_MATRIX_DATA(iD, Y+25,X+20, inWS) * (0.00433600f);
 result += GET_MATRIX_DATA(iD, Y+27,X+20, inWS) * (0.00823448f);
 result += GET_MATRIX_DATA(iD, Y+23,X+21, inWS) * (0.01948986f);
 result += GET_MATRIX_DATA(iD, Y+7,X+22, inWS) * (0.00422993f);
 result += GET_MATRIX_DATA(iD, Y+27,X+22, inWS) * (0.01551102f);
 result += GET_MATRIX_DATA(iD, Y+11,X+23, inWS) * (0.02991301f);
 result += GET_MATRIX_DATA(iD, Y+12,X+23, inWS) * (0.00299178f);
 result += GET_MATRIX_DATA(iD, Y+25,X+23, inWS) * (0.01056964f);
 result += GET_MATRIX_DATA(iD, Y+17,X+32, inWS) * (-0.01442667f);
 result += GET_MATRIX_DATA(iD, Y+21,X+32, inWS) * (-0.00061916f);
 result += GET_MATRIX_DATA(iD, Y+23,X+32, inWS) * (-0.00850284f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00474000f);
desc[10] = (result > (56.42f));

 /***********************************************
                  Projection: 12
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+28,X+3, inWS) * (-0.00423337f);
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (-0.00736552f);
 result += GET_MATRIX_DATA(iD, Y+14,X+7, inWS) * (0.00414316f);
 result += GET_MATRIX_DATA(iD, Y+16,X+7, inWS) * (0.02127144f);
 result += GET_MATRIX_DATA(iD, Y+17,X+7, inWS) * (0.00515932f);
 result += GET_MATRIX_DATA(iD, Y+21,X+8, inWS) * (0.01916452f);
 result += GET_MATRIX_DATA(iD, Y+30,X+8, inWS) * (-0.00185337f);
 result += GET_MATRIX_DATA(iD, Y+15,X+9, inWS) * (0.02507469f);
 result += GET_MATRIX_DATA(iD, Y+17,X+10, inWS) * (0.00972278f);
 result += GET_MATRIX_DATA(iD, Y+19,X+10, inWS) * (0.00779811f);
 result += GET_MATRIX_DATA(iD, Y+20,X+10, inWS) * (0.00552609f);
 result += GET_MATRIX_DATA(iD, Y+21,X+10, inWS) * (0.00965004f);
 result += GET_MATRIX_DATA(iD, Y+17,X+11, inWS) * (0.04859753f);
 result += GET_MATRIX_DATA(iD, Y+3,X+12, inWS) * (-0.00549664f);
 result += GET_MATRIX_DATA(iD, Y+4,X+12, inWS) * (-0.02936389f);
 result += GET_MATRIX_DATA(iD, Y+15,X+12, inWS) * (0.01166190f);
 result += GET_MATRIX_DATA(iD, Y+19,X+12, inWS) * (0.01281022f);
 result += GET_MATRIX_DATA(iD, Y+6,X+14, inWS) * (-0.01368139f);
 result += GET_MATRIX_DATA(iD, Y+8,X+14, inWS) * (-0.08293561f);
 result += GET_MATRIX_DATA(iD, Y+10,X+15, inWS) * (-0.01549007f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (0.00988885f);
 result += GET_MATRIX_DATA(iD, Y+16,X+18, inWS) * (0.07513780f);
 result += GET_MATRIX_DATA(iD, Y+32,X+19, inWS) * (0.01414203f);
 result += GET_MATRIX_DATA(iD, Y+30,X+20, inWS) * (0.02068657f);
 result += GET_MATRIX_DATA(iD, Y+30,X+21, inWS) * (0.02161041f);
 result += GET_MATRIX_DATA(iD, Y+3,X+22, inWS) * (0.02321981f);
 result += GET_MATRIX_DATA(iD, Y+19,X+22, inWS) * (-0.05320795f);
 result += GET_MATRIX_DATA(iD, Y+21,X+22, inWS) * (-0.03667685f);
 result += GET_MATRIX_DATA(iD, Y+17,X+23, inWS) * (-0.00676849f);
 result += GET_MATRIX_DATA(iD, Y+19,X+23, inWS) * (-0.00215363f);
 result += GET_MATRIX_DATA(iD, Y+3,X+24, inWS) * (0.01292026f);
 result += GET_MATRIX_DATA(iD, Y+13,X+24, inWS) * (-0.00916913f);
 result += GET_MATRIX_DATA(iD, Y+14,X+24, inWS) * (-0.03979989f);
 result += GET_MATRIX_DATA(iD, Y+17,X+24, inWS) * (-0.02584122f);
 result += GET_MATRIX_DATA(iD, Y+20,X+24, inWS) * (-0.03599245f);
 result += GET_MATRIX_DATA(iD, Y+21,X+24, inWS) * (-0.01368042f);
 result += GET_MATRIX_DATA(iD, Y+23,X+24, inWS) * (-0.00371180f);
 result += GET_MATRIX_DATA(iD, Y+15,X+25, inWS) * (-0.02306297f);
 result += GET_MATRIX_DATA(iD, Y+17,X+25, inWS) * (-0.01934151f);
 result += GET_MATRIX_DATA(iD, Y+18,X+25, inWS) * (-0.02011403f);
 result += GET_MATRIX_DATA(iD, Y+21,X+25, inWS) * (-0.01660459f);
 result += GET_MATRIX_DATA(iD, Y+22,X+25, inWS) * (-0.00650629f);
 result += GET_MATRIX_DATA(iD, Y+30,X+25, inWS) * (0.01834411f);
 result += GET_MATRIX_DATA(iD, Y+3,X+26, inWS) * (0.03685174f);
 result += GET_MATRIX_DATA(iD, Y+13,X+26, inWS) * (-0.00161331f);
 result += GET_MATRIX_DATA(iD, Y+14,X+26, inWS) * (-0.00540723f);
 result += GET_MATRIX_DATA(iD, Y+22,X+26, inWS) * (-0.00409046f);
 result += GET_MATRIX_DATA(iD, Y+14,X+27, inWS) * (-0.01753477f);
 result += GET_MATRIX_DATA(iD, Y+16,X+27, inWS) * (-0.02398364f);
 result += GET_MATRIX_DATA(iD, Y+18,X+27, inWS) * (-0.02346313f);
 result += GET_MATRIX_DATA(iD, Y+20,X+27, inWS) * (-0.01712795f);
 result += GET_MATRIX_DATA(iD, Y+21,X+28, inWS) * (-0.01939692f);
 result += GET_MATRIX_DATA(iD, Y+15,X+29, inWS) * (-0.00776988f);
 result += GET_MATRIX_DATA(iD, Y+18,X+29, inWS) * (-0.01744371f);
 result += GET_MATRIX_DATA(iD, Y+1,X+30, inWS) * (0.02892786f);
 result += GET_MATRIX_DATA(iD, Y+5,X+30, inWS) * (0.02355019f);
 result += GET_MATRIX_DATA(iD, Y+30,X+31, inWS) * (0.01462480f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00443464f);
desc[11] = (result > (-79.28f));

 /***********************************************
                  Projection: 13
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+16,X+3, inWS) * (-0.00207773f);
 result += GET_MATRIX_DATA(iD, Y+18,X+4, inWS) * (-0.00353302f);
 result += GET_MATRIX_DATA(iD, Y+12,X+5, inWS) * (-0.00561291f);
 result += GET_MATRIX_DATA(iD, Y+14,X+5, inWS) * (-0.00197251f);
 result += GET_MATRIX_DATA(iD, Y+16,X+5, inWS) * (-0.01462497f);
 result += GET_MATRIX_DATA(iD, Y+12,X+7, inWS) * (-0.00443499f);
 result += GET_MATRIX_DATA(iD, Y+30,X+7, inWS) * (0.01181188f);
 result += GET_MATRIX_DATA(iD, Y+21,X+8, inWS) * (-0.01675723f);
 result += GET_MATRIX_DATA(iD, Y+3,X+9, inWS) * (0.00181857f);
 result += GET_MATRIX_DATA(iD, Y+13,X+9, inWS) * (-0.00249149f);
 result += GET_MATRIX_DATA(iD, Y+2,X+10, inWS) * (0.00638987f);
 result += GET_MATRIX_DATA(iD, Y+30,X+10, inWS) * (0.02083096f);
 result += GET_MATRIX_DATA(iD, Y+3,X+11, inWS) * (0.01049367f);
 result += GET_MATRIX_DATA(iD, Y+22,X+11, inWS) * (-0.01692617f);
 result += GET_MATRIX_DATA(iD, Y+30,X+11, inWS) * (0.01042099f);
 result += GET_MATRIX_DATA(iD, Y+32,X+11, inWS) * (0.01000018f);
 result += GET_MATRIX_DATA(iD, Y+30,X+12, inWS) * (0.00973333f);
 result += GET_MATRIX_DATA(iD, Y+3,X+13, inWS) * (0.00856530f);
 result += GET_MATRIX_DATA(iD, Y+16,X+13, inWS) * (0.10085327f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (0.06155069f);
 result += GET_MATRIX_DATA(iD, Y+10,X+14, inWS) * (-0.00949046f);
 result += GET_MATRIX_DATA(iD, Y+11,X+14, inWS) * (-0.00938756f);
 result += GET_MATRIX_DATA(iD, Y+17,X+14, inWS) * (0.05468461f);
 result += GET_MATRIX_DATA(iD, Y+24,X+14, inWS) * (-0.00599215f);
 result += GET_MATRIX_DATA(iD, Y+32,X+14, inWS) * (0.01062773f);
 result += GET_MATRIX_DATA(iD, Y+10,X+15, inWS) * (-0.01412042f);
 result += GET_MATRIX_DATA(iD, Y+22,X+15, inWS) * (-0.07123752f);
 result += GET_MATRIX_DATA(iD, Y+24,X+15, inWS) * (-0.00364288f);
 result += GET_MATRIX_DATA(iD, Y+1,X+16, inWS) * (0.00133552f);
 result += GET_MATRIX_DATA(iD, Y+10,X+16, inWS) * (-0.03757261f);
 result += GET_MATRIX_DATA(iD, Y+11,X+16, inWS) * (-0.04090237f);
 result += GET_MATRIX_DATA(iD, Y+12,X+16, inWS) * (-0.04446184f);
 result += GET_MATRIX_DATA(iD, Y+13,X+16, inWS) * (-0.00785211f);
 result += GET_MATRIX_DATA(iD, Y+22,X+16, inWS) * (-0.02007398f);
 result += GET_MATRIX_DATA(iD, Y+24,X+16, inWS) * (-0.04319874f);
 result += GET_MATRIX_DATA(iD, Y+32,X+16, inWS) * (0.00519396f);
 result += GET_MATRIX_DATA(iD, Y+11,X+17, inWS) * (-0.01686031f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (-0.01847926f);
 result += GET_MATRIX_DATA(iD, Y+20,X+17, inWS) * (-0.02193355f);
 result += GET_MATRIX_DATA(iD, Y+22,X+17, inWS) * (-0.01105438f);
 result += GET_MATRIX_DATA(iD, Y+24,X+17, inWS) * (-0.01413145f);
 result += GET_MATRIX_DATA(iD, Y+2,X+18, inWS) * (0.00713910f);
 result += GET_MATRIX_DATA(iD, Y+10,X+18, inWS) * (-0.05511850f);
 result += GET_MATRIX_DATA(iD, Y+11,X+18, inWS) * (-0.01294126f);
 result += GET_MATRIX_DATA(iD, Y+22,X+18, inWS) * (-0.00786369f);
 result += GET_MATRIX_DATA(iD, Y+24,X+18, inWS) * (-0.02704075f);
 result += GET_MATRIX_DATA(iD, Y+25,X+18, inWS) * (-0.01177911f);
 result += GET_MATRIX_DATA(iD, Y+17,X+19, inWS) * (0.05026084f);
 result += GET_MATRIX_DATA(iD, Y+18,X+19, inWS) * (0.02586058f);
 result += GET_MATRIX_DATA(iD, Y+3,X+20, inWS) * (0.00406741f);
 result += GET_MATRIX_DATA(iD, Y+15,X+20, inWS) * (0.06161454f);
 result += GET_MATRIX_DATA(iD, Y+16,X+20, inWS) * (0.05230979f);
 result += GET_MATRIX_DATA(iD, Y+24,X+20, inWS) * (-0.03662097f);
 result += GET_MATRIX_DATA(iD, Y+16,X+21, inWS) * (0.04659852f);
 result += GET_MATRIX_DATA(iD, Y+17,X+21, inWS) * (0.03683335f);
 result += GET_MATRIX_DATA(iD, Y+18,X+21, inWS) * (0.04489199f);
 result += GET_MATRIX_DATA(iD, Y+25,X+21, inWS) * (-0.00409191f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00252672f);
desc[12] = (result > (437.75f));

 /***********************************************
                  Projection: 14
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+25,X+1, inWS) * (0.00056434f);
 result += GET_MATRIX_DATA(iD, Y+3,X+2, inWS) * (-0.00947423f);
 result += GET_MATRIX_DATA(iD, Y+27,X+3, inWS) * (0.00648244f);
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (0.01537152f);
 result += GET_MATRIX_DATA(iD, Y+31,X+3, inWS) * (0.00314115f);
 result += GET_MATRIX_DATA(iD, Y+3,X+5, inWS) * (-0.01184609f);
 result += GET_MATRIX_DATA(iD, Y+22,X+10, inWS) * (-0.00141169f);
 result += GET_MATRIX_DATA(iD, Y+23,X+10, inWS) * (-0.00436350f);
 result += GET_MATRIX_DATA(iD, Y+13,X+11, inWS) * (0.02238108f);
 result += GET_MATRIX_DATA(iD, Y+11,X+12, inWS) * (0.00779814f);
 result += GET_MATRIX_DATA(iD, Y+14,X+12, inWS) * (0.01688690f);
 result += GET_MATRIX_DATA(iD, Y+22,X+12, inWS) * (-0.02893881f);
 result += GET_MATRIX_DATA(iD, Y+12,X+13, inWS) * (0.04661969f);
 result += GET_MATRIX_DATA(iD, Y+14,X+13, inWS) * (0.04875958f);
 result += GET_MATRIX_DATA(iD, Y+15,X+13, inWS) * (0.02021598f);
 result += GET_MATRIX_DATA(iD, Y+23,X+13, inWS) * (-0.04418069f);
 result += GET_MATRIX_DATA(iD, Y+13,X+14, inWS) * (0.07985060f);
 result += GET_MATRIX_DATA(iD, Y+14,X+14, inWS) * (0.00174658f);
 result += GET_MATRIX_DATA(iD, Y+20,X+14, inWS) * (-0.07526196f);
 result += GET_MATRIX_DATA(iD, Y+21,X+14, inWS) * (-0.02441685f);
 result += GET_MATRIX_DATA(iD, Y+22,X+14, inWS) * (-0.00780286f);
 result += GET_MATRIX_DATA(iD, Y+14,X+15, inWS) * (0.00095213f);
 result += GET_MATRIX_DATA(iD, Y+22,X+15, inWS) * (-0.02743291f);
 result += GET_MATRIX_DATA(iD, Y+9,X+18, inWS) * (-0.00576630f);
 result += GET_MATRIX_DATA(iD, Y+9,X+19, inWS) * (-0.01251534f);
 result += GET_MATRIX_DATA(iD, Y+21,X+19, inWS) * (0.01818004f);
 result += GET_MATRIX_DATA(iD, Y+24,X+19, inWS) * (0.00313042f);
 result += GET_MATRIX_DATA(iD, Y+25,X+19, inWS) * (0.00101002f);
 result += GET_MATRIX_DATA(iD, Y+10,X+20, inWS) * (-0.00792038f);
 result += GET_MATRIX_DATA(iD, Y+11,X+20, inWS) * (-0.01669971f);
 result += GET_MATRIX_DATA(iD, Y+13,X+20, inWS) * (-0.10291196f);
 result += GET_MATRIX_DATA(iD, Y+14,X+20, inWS) * (-0.00041381f);
 result += GET_MATRIX_DATA(iD, Y+15,X+20, inWS) * (-0.00638835f);
 result += GET_MATRIX_DATA(iD, Y+20,X+20, inWS) * (0.07214631f);
 result += GET_MATRIX_DATA(iD, Y+21,X+20, inWS) * (0.02689924f);
 result += GET_MATRIX_DATA(iD, Y+23,X+20, inWS) * (0.02986895f);
 result += GET_MATRIX_DATA(iD, Y+25,X+20, inWS) * (0.01572017f);
 result += GET_MATRIX_DATA(iD, Y+13,X+21, inWS) * (-0.02190096f);
 result += GET_MATRIX_DATA(iD, Y+14,X+21, inWS) * (-0.02497592f);
 result += GET_MATRIX_DATA(iD, Y+21,X+21, inWS) * (0.06700865f);
 result += GET_MATRIX_DATA(iD, Y+25,X+21, inWS) * (0.00122677f);
 result += GET_MATRIX_DATA(iD, Y+26,X+21, inWS) * (0.00977616f);
 result += GET_MATRIX_DATA(iD, Y+12,X+22, inWS) * (-0.00904134f);
 result += GET_MATRIX_DATA(iD, Y+21,X+22, inWS) * (0.00311695f);
 result += GET_MATRIX_DATA(iD, Y+22,X+22, inWS) * (0.01766615f);
 result += GET_MATRIX_DATA(iD, Y+25,X+22, inWS) * (0.00598082f);
 result += GET_MATRIX_DATA(iD, Y+23,X+23, inWS) * (0.00911781f);
 result += GET_MATRIX_DATA(iD, Y+2,X+26, inWS) * (0.01627937f);
 result += GET_MATRIX_DATA(iD, Y+3,X+30, inWS) * (0.02202108f);
 result += GET_MATRIX_DATA(iD, Y+6,X+30, inWS) * (0.01247968f);
 result += GET_MATRIX_DATA(iD, Y+7,X+30, inWS) * (0.00214967f);
 result += GET_MATRIX_DATA(iD, Y+18,X+30, inWS) * (-0.00062646f);
 result += GET_MATRIX_DATA(iD, Y+19,X+30, inWS) * (-0.00001772f);
 result += GET_MATRIX_DATA(iD, Y+24,X+30, inWS) * (-0.00383454f);
 result += GET_MATRIX_DATA(iD, Y+27,X+30, inWS) * (-0.00294475f);
 result += GET_MATRIX_DATA(iD, Y+30,X+30, inWS) * (-0.00381522f);
 result += GET_MATRIX_DATA(iD, Y+30,X+31, inWS) * (-0.00705208f);
 result += GET_MATRIX_DATA(iD, Y+23,X+32, inWS) * (-0.00631082f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00276900f);
desc[13] = (result > (76.31f));

 /***********************************************
                  Projection: 15
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+13,X+3, inWS) * (-0.00804493f);
 result += GET_MATRIX_DATA(iD, Y+15,X+3, inWS) * (-0.02701164f);
 result += GET_MATRIX_DATA(iD, Y+1,X+4, inWS) * (0.00439985f);
 result += GET_MATRIX_DATA(iD, Y+1,X+5, inWS) * (0.00407408f);
 result += GET_MATRIX_DATA(iD, Y+13,X+5, inWS) * (-0.00147267f);
 result += GET_MATRIX_DATA(iD, Y+25,X+6, inWS) * (0.00433105f);
 result += GET_MATRIX_DATA(iD, Y+1,X+7, inWS) * (0.00060531f);
 result += GET_MATRIX_DATA(iD, Y+22,X+7, inWS) * (0.01321616f);
 result += GET_MATRIX_DATA(iD, Y+12,X+8, inWS) * (-0.00189792f);
 result += GET_MATRIX_DATA(iD, Y+22,X+8, inWS) * (0.01644679f);
 result += GET_MATRIX_DATA(iD, Y+24,X+8, inWS) * (0.00673832f);
 result += GET_MATRIX_DATA(iD, Y+19,X+9, inWS) * (0.00943975f);
 result += GET_MATRIX_DATA(iD, Y+21,X+9, inWS) * (0.00462882f);
 result += GET_MATRIX_DATA(iD, Y+22,X+9, inWS) * (0.00294445f);
 result += GET_MATRIX_DATA(iD, Y+24,X+9, inWS) * (0.00292908f);
 result += GET_MATRIX_DATA(iD, Y+2,X+10, inWS) * (0.01964967f);
 result += GET_MATRIX_DATA(iD, Y+11,X+10, inWS) * (-0.03823412f);
 result += GET_MATRIX_DATA(iD, Y+12,X+10, inWS) * (-0.01349997f);
 result += GET_MATRIX_DATA(iD, Y+20,X+10, inWS) * (0.00826645f);
 result += GET_MATRIX_DATA(iD, Y+21,X+10, inWS) * (0.03109380f);
 result += GET_MATRIX_DATA(iD, Y+23,X+10, inWS) * (0.02576693f);
 result += GET_MATRIX_DATA(iD, Y+24,X+10, inWS) * (0.01888992f);
 result += GET_MATRIX_DATA(iD, Y+10,X+11, inWS) * (-0.00018226f);
 result += GET_MATRIX_DATA(iD, Y+21,X+11, inWS) * (0.01732331f);
 result += GET_MATRIX_DATA(iD, Y+22,X+11, inWS) * (0.01222537f);
 result += GET_MATRIX_DATA(iD, Y+2,X+13, inWS) * (0.01234586f);
 result += GET_MATRIX_DATA(iD, Y+16,X+13, inWS) * (-0.02587342f);
 result += GET_MATRIX_DATA(iD, Y+1,X+14, inWS) * (0.00010317f);
 result += GET_MATRIX_DATA(iD, Y+18,X+14, inWS) * (-0.00303995f);
 result += GET_MATRIX_DATA(iD, Y+28,X+15, inWS) * (-0.01092666f);
 result += GET_MATRIX_DATA(iD, Y+30,X+15, inWS) * (-0.02522489f);
 result += GET_MATRIX_DATA(iD, Y+1,X+16, inWS) * (0.01790422f);
 result += GET_MATRIX_DATA(iD, Y+3,X+16, inWS) * (0.00425759f);
 result += GET_MATRIX_DATA(iD, Y+13,X+16, inWS) * (0.07811986f);
 result += GET_MATRIX_DATA(iD, Y+30,X+16, inWS) * (-0.00319826f);
 result += GET_MATRIX_DATA(iD, Y+14,X+17, inWS) * (0.08807727f);
 result += GET_MATRIX_DATA(iD, Y+19,X+17, inWS) * (-0.01444015f);
 result += GET_MATRIX_DATA(iD, Y+22,X+17, inWS) * (-0.01993653f);
 result += GET_MATRIX_DATA(iD, Y+28,X+17, inWS) * (-0.02666211f);
 result += GET_MATRIX_DATA(iD, Y+32,X+17, inWS) * (-0.00571117f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (0.03533502f);
 result += GET_MATRIX_DATA(iD, Y+1,X+19, inWS) * (0.01766663f);
 result += GET_MATRIX_DATA(iD, Y+9,X+19, inWS) * (-0.00017518f);
 result += GET_MATRIX_DATA(iD, Y+9,X+20, inWS) * (-0.00735103f);
 result += GET_MATRIX_DATA(iD, Y+10,X+20, inWS) * (-0.00210077f);
 result += GET_MATRIX_DATA(iD, Y+11,X+21, inWS) * (-0.06275056f);
 result += GET_MATRIX_DATA(iD, Y+12,X+21, inWS) * (-0.00279230f);
 result += GET_MATRIX_DATA(iD, Y+9,X+22, inWS) * (-0.01944952f);
 result += GET_MATRIX_DATA(iD, Y+11,X+22, inWS) * (-0.00287958f);
 result += GET_MATRIX_DATA(iD, Y+12,X+23, inWS) * (-0.02295429f);
 result += GET_MATRIX_DATA(iD, Y+13,X+23, inWS) * (-0.00071107f);
 result += GET_MATRIX_DATA(iD, Y+23,X+23, inWS) * (0.01383235f);
 result += GET_MATRIX_DATA(iD, Y+20,X+24, inWS) * (0.00439637f);
 result += GET_MATRIX_DATA(iD, Y+21,X+24, inWS) * (0.03186888f);
 result += GET_MATRIX_DATA(iD, Y+22,X+24, inWS) * (0.01013701f);
 result += GET_MATRIX_DATA(iD, Y+24,X+24, inWS) * (0.00218384f);
 result += GET_MATRIX_DATA(iD, Y+3,X+25, inWS) * (0.00045824f);
 result += GET_MATRIX_DATA(iD, Y+21,X+25, inWS) * (0.01109914f);
 result += GET_MATRIX_DATA(iD, Y+22,X+25, inWS) * (0.00137976f);
 result += GET_MATRIX_DATA(iD, Y+24,X+25, inWS) * (0.00731296f);
 result += GET_MATRIX_DATA(iD, Y+11,X+26, inWS) * (-0.00084908f);
 result += GET_MATRIX_DATA(iD, Y+13,X+26, inWS) * (-0.00694794f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00387127f);
desc[14] = (result > (-20.55f));

 /***********************************************
                  Projection: 16
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+17,X+1, inWS) * (-0.00743209f);
 result += GET_MATRIX_DATA(iD, Y+30,X+7, inWS) * (0.01257558f);
 result += GET_MATRIX_DATA(iD, Y+7,X+8, inWS) * (0.00482707f);
 result += GET_MATRIX_DATA(iD, Y+11,X+9, inWS) * (0.01377361f);
 result += GET_MATRIX_DATA(iD, Y+13,X+9, inWS) * (0.00500543f);
 result += GET_MATRIX_DATA(iD, Y+30,X+9, inWS) * (0.01038816f);
 result += GET_MATRIX_DATA(iD, Y+11,X+10, inWS) * (0.00352822f);
 result += GET_MATRIX_DATA(iD, Y+13,X+10, inWS) * (0.01409199f);
 result += GET_MATRIX_DATA(iD, Y+14,X+10, inWS) * (0.00127689f);
 result += GET_MATRIX_DATA(iD, Y+32,X+11, inWS) * (0.00623800f);
 result += GET_MATRIX_DATA(iD, Y+19,X+12, inWS) * (-0.00371870f);
 result += GET_MATRIX_DATA(iD, Y+22,X+12, inWS) * (-0.03442728f);
 result += GET_MATRIX_DATA(iD, Y+23,X+12, inWS) * (-0.00338115f);
 result += GET_MATRIX_DATA(iD, Y+16,X+13, inWS) * (-0.01212786f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (-0.06211412f);
 result += GET_MATRIX_DATA(iD, Y+18,X+13, inWS) * (-0.06963125f);
 result += GET_MATRIX_DATA(iD, Y+23,X+13, inWS) * (-0.01955595f);
 result += GET_MATRIX_DATA(iD, Y+5,X+14, inWS) * (-0.00980274f);
 result += GET_MATRIX_DATA(iD, Y+24,X+14, inWS) * (-0.01935826f);
 result += GET_MATRIX_DATA(iD, Y+13,X+15, inWS) * (0.04459039f);
 result += GET_MATRIX_DATA(iD, Y+6,X+16, inWS) * (-0.00668106f);
 result += GET_MATRIX_DATA(iD, Y+7,X+16, inWS) * (-0.00877542f);
 result += GET_MATRIX_DATA(iD, Y+13,X+16, inWS) * (0.01327323f);
 result += GET_MATRIX_DATA(iD, Y+14,X+16, inWS) * (0.00154760f);
 result += GET_MATRIX_DATA(iD, Y+20,X+16, inWS) * (0.17466532f);
 result += GET_MATRIX_DATA(iD, Y+3,X+17, inWS) * (-0.01147473f);
 result += GET_MATRIX_DATA(iD, Y+4,X+17, inWS) * (-0.00059933f);
 result += GET_MATRIX_DATA(iD, Y+18,X+17, inWS) * (0.00588197f);
 result += GET_MATRIX_DATA(iD, Y+19,X+17, inWS) * (0.06861818f);
 result += GET_MATRIX_DATA(iD, Y+20,X+17, inWS) * (0.02619034f);
 result += GET_MATRIX_DATA(iD, Y+25,X+17, inWS) * (-0.01242774f);
 result += GET_MATRIX_DATA(iD, Y+6,X+18, inWS) * (-0.00206283f);
 result += GET_MATRIX_DATA(iD, Y+7,X+18, inWS) * (-0.00364997f);
 result += GET_MATRIX_DATA(iD, Y+10,X+18, inWS) * (-0.01728762f);
 result += GET_MATRIX_DATA(iD, Y+25,X+18, inWS) * (-0.01381309f);
 result += GET_MATRIX_DATA(iD, Y+32,X+18, inWS) * (0.00328952f);
 result += GET_MATRIX_DATA(iD, Y+6,X+19, inWS) * (-0.00923119f);
 result += GET_MATRIX_DATA(iD, Y+7,X+19, inWS) * (-0.00421767f);
 result += GET_MATRIX_DATA(iD, Y+16,X+19, inWS) * (-0.02421821f);
 result += GET_MATRIX_DATA(iD, Y+23,X+19, inWS) * (-0.02445744f);
 result += GET_MATRIX_DATA(iD, Y+25,X+19, inWS) * (-0.01449077f);
 result += GET_MATRIX_DATA(iD, Y+16,X+20, inWS) * (-0.01457243f);
 result += GET_MATRIX_DATA(iD, Y+22,X+20, inWS) * (-0.02813260f);
 result += GET_MATRIX_DATA(iD, Y+23,X+20, inWS) * (-0.01083983f);
 result += GET_MATRIX_DATA(iD, Y+24,X+20, inWS) * (-0.01128794f);
 result += GET_MATRIX_DATA(iD, Y+25,X+21, inWS) * (-0.00796120f);
 result += GET_MATRIX_DATA(iD, Y+10,X+23, inWS) * (0.00667267f);
 result += GET_MATRIX_DATA(iD, Y+13,X+23, inWS) * (0.01090036f);
 result += GET_MATRIX_DATA(iD, Y+14,X+23, inWS) * (0.01561920f);
 result += GET_MATRIX_DATA(iD, Y+12,X+24, inWS) * (0.03464707f);
 result += GET_MATRIX_DATA(iD, Y+14,X+24, inWS) * (0.00337261f);
 result += GET_MATRIX_DATA(iD, Y+15,X+24, inWS) * (0.04360510f);
 result += GET_MATRIX_DATA(iD, Y+17,X+24, inWS) * (0.02144062f);
 result += GET_MATRIX_DATA(iD, Y+31,X+24, inWS) * (0.01163850f);
 result += GET_MATRIX_DATA(iD, Y+10,X+25, inWS) * (0.02162433f);
 result += GET_MATRIX_DATA(iD, Y+12,X+25, inWS) * (0.00045042f);
 result += GET_MATRIX_DATA(iD, Y+14,X+25, inWS) * (0.00212654f);
 result += GET_MATRIX_DATA(iD, Y+30,X+27, inWS) * (0.00439903f);
 result += GET_MATRIX_DATA(iD, Y+17,X+31, inWS) * (-0.00131620f);
 result += GET_MATRIX_DATA(iD, Y+19,X+31, inWS) * (-0.00602669f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00228253f);
desc[15] = (result > (110.59f));

 /***********************************************
                  Projection: 17
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+21,X+1, inWS) * (-0.01627627f);
 result += GET_MATRIX_DATA(iD, Y+3,X+2, inWS) * (0.01907936f);
 result += GET_MATRIX_DATA(iD, Y+17,X+2, inWS) * (-0.00385730f);
 result += GET_MATRIX_DATA(iD, Y+5,X+3, inWS) * (0.02301206f);
 result += GET_MATRIX_DATA(iD, Y+7,X+3, inWS) * (0.00063566f);
 result += GET_MATRIX_DATA(iD, Y+18,X+3, inWS) * (-0.01199778f);
 result += GET_MATRIX_DATA(iD, Y+20,X+3, inWS) * (-0.00481221f);
 result += GET_MATRIX_DATA(iD, Y+23,X+3, inWS) * (-0.02358225f);
 result += GET_MATRIX_DATA(iD, Y+3,X+5, inWS) * (0.00632686f);
 result += GET_MATRIX_DATA(iD, Y+16,X+5, inWS) * (-0.00125748f);
 result += GET_MATRIX_DATA(iD, Y+18,X+10, inWS) * (0.00584555f);
 result += GET_MATRIX_DATA(iD, Y+19,X+10, inWS) * (0.00450514f);
 result += GET_MATRIX_DATA(iD, Y+24,X+10, inWS) * (0.03943949f);
 result += GET_MATRIX_DATA(iD, Y+28,X+10, inWS) * (0.00333751f);
 result += GET_MATRIX_DATA(iD, Y+7,X+11, inWS) * (-0.01558916f);
 result += GET_MATRIX_DATA(iD, Y+8,X+11, inWS) * (-0.01923109f);
 result += GET_MATRIX_DATA(iD, Y+9,X+11, inWS) * (-0.01422881f);
 result += GET_MATRIX_DATA(iD, Y+4,X+12, inWS) * (-0.00129951f);
 result += GET_MATRIX_DATA(iD, Y+15,X+12, inWS) * (0.01649361f);
 result += GET_MATRIX_DATA(iD, Y+16,X+12, inWS) * (0.03488129f);
 result += GET_MATRIX_DATA(iD, Y+25,X+12, inWS) * (0.00505210f);
 result += GET_MATRIX_DATA(iD, Y+6,X+13, inWS) * (-0.00209555f);
 result += GET_MATRIX_DATA(iD, Y+7,X+13, inWS) * (-0.02084084f);
 result += GET_MATRIX_DATA(iD, Y+8,X+13, inWS) * (-0.00368901f);
 result += GET_MATRIX_DATA(iD, Y+9,X+13, inWS) * (-0.00957928f);
 result += GET_MATRIX_DATA(iD, Y+10,X+13, inWS) * (-0.00605395f);
 result += GET_MATRIX_DATA(iD, Y+14,X+14, inWS) * (0.05894418f);
 result += GET_MATRIX_DATA(iD, Y+18,X+14, inWS) * (-0.02939509f);
 result += GET_MATRIX_DATA(iD, Y+14,X+15, inWS) * (0.03161170f);
 result += GET_MATRIX_DATA(iD, Y+19,X+15, inWS) * (-0.06969409f);
 result += GET_MATRIX_DATA(iD, Y+20,X+15, inWS) * (-0.08158775f);
 result += GET_MATRIX_DATA(iD, Y+19,X+18, inWS) * (0.04396143f);
 result += GET_MATRIX_DATA(iD, Y+14,X+19, inWS) * (-0.00722631f);
 result += GET_MATRIX_DATA(iD, Y+15,X+19, inWS) * (-0.06608641f);
 result += GET_MATRIX_DATA(iD, Y+19,X+19, inWS) * (0.08352881f);
 result += GET_MATRIX_DATA(iD, Y+20,X+19, inWS) * (0.04435675f);
 result += GET_MATRIX_DATA(iD, Y+8,X+20, inWS) * (0.04445659f);
 result += GET_MATRIX_DATA(iD, Y+10,X+20, inWS) * (0.00088577f);
 result += GET_MATRIX_DATA(iD, Y+15,X+20, inWS) * (-0.10102992f);
 result += GET_MATRIX_DATA(iD, Y+32,X+20, inWS) * (-0.00409105f);
 result += GET_MATRIX_DATA(iD, Y+6,X+21, inWS) * (0.00076435f);
 result += GET_MATRIX_DATA(iD, Y+9,X+21, inWS) * (0.00386529f);
 result += GET_MATRIX_DATA(iD, Y+10,X+21, inWS) * (0.01934320f);
 result += GET_MATRIX_DATA(iD, Y+9,X+22, inWS) * (0.01952482f);
 result += GET_MATRIX_DATA(iD, Y+16,X+22, inWS) * (-0.00931494f);
 result += GET_MATRIX_DATA(iD, Y+17,X+22, inWS) * (-0.01639027f);
 result += GET_MATRIX_DATA(iD, Y+30,X+22, inWS) * (-0.01437151f);
 result += GET_MATRIX_DATA(iD, Y+2,X+27, inWS) * (-0.00043610f);
 result += GET_MATRIX_DATA(iD, Y+3,X+30, inWS) * (-0.02270551f);
 result += GET_MATRIX_DATA(iD, Y+5,X+30, inWS) * (-0.00745890f);
 result += GET_MATRIX_DATA(iD, Y+18,X+30, inWS) * (0.00947238f);
 result += GET_MATRIX_DATA(iD, Y+19,X+30, inWS) * (0.00482734f);
 result += GET_MATRIX_DATA(iD, Y+19,X+32, inWS) * (0.00422500f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00120267f);
desc[16] = (result > (-14.04f));

 /***********************************************
                  Projection: 18
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+13,X+2, inWS) * (0.01626970f);
 result += GET_MATRIX_DATA(iD, Y+3,X+3, inWS) * (-0.00137471f);
 result += GET_MATRIX_DATA(iD, Y+13,X+3, inWS) * (0.00513846f);
 result += GET_MATRIX_DATA(iD, Y+15,X+3, inWS) * (0.03493449f);
 result += GET_MATRIX_DATA(iD, Y+16,X+3, inWS) * (0.00812044f);
 result += GET_MATRIX_DATA(iD, Y+18,X+3, inWS) * (0.01507850f);
 result += GET_MATRIX_DATA(iD, Y+19,X+3, inWS) * (0.00999401f);
 result += GET_MATRIX_DATA(iD, Y+20,X+3, inWS) * (0.01106663f);
 result += GET_MATRIX_DATA(iD, Y+28,X+3, inWS) * (-0.00475087f);
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (-0.02510491f);
 result += GET_MATRIX_DATA(iD, Y+1,X+4, inWS) * (-0.00500361f);
 result += GET_MATRIX_DATA(iD, Y+3,X+5, inWS) * (-0.00506081f);
 result += GET_MATRIX_DATA(iD, Y+13,X+5, inWS) * (0.01027240f);
 result += GET_MATRIX_DATA(iD, Y+17,X+5, inWS) * (0.00544592f);
 result += GET_MATRIX_DATA(iD, Y+19,X+5, inWS) * (0.02165908f);
 result += GET_MATRIX_DATA(iD, Y+30,X+5, inWS) * (-0.01396648f);
 result += GET_MATRIX_DATA(iD, Y+21,X+7, inWS) * (0.00256772f);
 result += GET_MATRIX_DATA(iD, Y+30,X+7, inWS) * (-0.01274890f);
 result += GET_MATRIX_DATA(iD, Y+32,X+7, inWS) * (-0.00290442f);
 result += GET_MATRIX_DATA(iD, Y+2,X+8, inWS) * (-0.01671544f);
 result += GET_MATRIX_DATA(iD, Y+6,X+8, inWS) * (-0.01172606f);
 result += GET_MATRIX_DATA(iD, Y+6,X+9, inWS) * (-0.00554125f);
 result += GET_MATRIX_DATA(iD, Y+10,X+10, inWS) * (-0.00766188f);
 result += GET_MATRIX_DATA(iD, Y+13,X+10, inWS) * (-0.01224251f);
 result += GET_MATRIX_DATA(iD, Y+16,X+10, inWS) * (-0.00079563f);
 result += GET_MATRIX_DATA(iD, Y+17,X+10, inWS) * (-0.00427591f);
 result += GET_MATRIX_DATA(iD, Y+8,X+11, inWS) * (-0.00001692f);
 result += GET_MATRIX_DATA(iD, Y+14,X+11, inWS) * (-0.03144683f);
 result += GET_MATRIX_DATA(iD, Y+17,X+11, inWS) * (-0.00496887f);
 result += GET_MATRIX_DATA(iD, Y+18,X+11, inWS) * (-0.00038369f);
 result += GET_MATRIX_DATA(iD, Y+11,X+13, inWS) * (-0.03218542f);
 result += GET_MATRIX_DATA(iD, Y+19,X+13, inWS) * (-0.01250929f);
 result += GET_MATRIX_DATA(iD, Y+20,X+13, inWS) * (-0.02364411f);
 result += GET_MATRIX_DATA(iD, Y+27,X+14, inWS) * (0.00060857f);
 result += GET_MATRIX_DATA(iD, Y+28,X+14, inWS) * (0.00799437f);
 result += GET_MATRIX_DATA(iD, Y+1,X+15, inWS) * (0.00022683f);
 result += GET_MATRIX_DATA(iD, Y+3,X+15, inWS) * (0.00261310f);
 result += GET_MATRIX_DATA(iD, Y+15,X+15, inWS) * (0.07353588f);
 result += GET_MATRIX_DATA(iD, Y+27,X+15, inWS) * (0.01550512f);
 result += GET_MATRIX_DATA(iD, Y+28,X+15, inWS) * (0.02615143f);
 result += GET_MATRIX_DATA(iD, Y+3,X+16, inWS) * (0.01233205f);
 result += GET_MATRIX_DATA(iD, Y+5,X+16, inWS) * (0.04008731f);
 result += GET_MATRIX_DATA(iD, Y+15,X+16, inWS) * (0.05712105f);
 result += GET_MATRIX_DATA(iD, Y+1,X+17, inWS) * (0.01671235f);
 result += GET_MATRIX_DATA(iD, Y+3,X+17, inWS) * (0.00126925f);
 result += GET_MATRIX_DATA(iD, Y+23,X+17, inWS) * (-0.01205225f);
 result += GET_MATRIX_DATA(iD, Y+3,X+18, inWS) * (0.00239203f);
 result += GET_MATRIX_DATA(iD, Y+5,X+18, inWS) * (0.01060587f);
 result += GET_MATRIX_DATA(iD, Y+10,X+18, inWS) * (-0.01639571f);
 result += GET_MATRIX_DATA(iD, Y+23,X+18, inWS) * (-0.00352760f);
 result += GET_MATRIX_DATA(iD, Y+3,X+19, inWS) * (0.02195389f);
 result += GET_MATRIX_DATA(iD, Y+11,X+19, inWS) * (-0.03474696f);
 result += GET_MATRIX_DATA(iD, Y+13,X+19, inWS) * (-0.01380888f);
 result += GET_MATRIX_DATA(iD, Y+23,X+19, inWS) * (-0.03191341f);
 result += GET_MATRIX_DATA(iD, Y+3,X+20, inWS) * (0.00889652f);
 result += GET_MATRIX_DATA(iD, Y+12,X+20, inWS) * (-0.02591889f);
 result += GET_MATRIX_DATA(iD, Y+14,X+20, inWS) * (-0.01769582f);
 result += GET_MATRIX_DATA(iD, Y+23,X+20, inWS) * (-0.01309937f);
 result += GET_MATRIX_DATA(iD, Y+3,X+21, inWS) * (0.00149144f);
 result += GET_MATRIX_DATA(iD, Y+3,X+22, inWS) * (0.00708617f);
 result += GET_MATRIX_DATA(iD, Y+17,X+22, inWS) * (0.01862816f);
 result += GET_MATRIX_DATA(iD, Y+19,X+25, inWS) * (0.00839129f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00193917f);
desc[17] = (result > (-57.89f));

 /***********************************************
                  Projection: 19
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+7,X+1, inWS) * (0.00151442f);
 result += GET_MATRIX_DATA(iD, Y+8,X+1, inWS) * (0.01265705f);
 result += GET_MATRIX_DATA(iD, Y+10,X+1, inWS) * (0.00157188f);
 result += GET_MATRIX_DATA(iD, Y+11,X+1, inWS) * (0.00360409f);
 result += GET_MATRIX_DATA(iD, Y+12,X+1, inWS) * (0.02548282f);
 result += GET_MATRIX_DATA(iD, Y+16,X+1, inWS) * (0.01171743f);
 result += GET_MATRIX_DATA(iD, Y+20,X+1, inWS) * (0.01428647f);
 result += GET_MATRIX_DATA(iD, Y+7,X+2, inWS) * (0.00262602f);
 result += GET_MATRIX_DATA(iD, Y+15,X+6, inWS) * (-0.00721421f);
 result += GET_MATRIX_DATA(iD, Y+17,X+6, inWS) * (-0.00346027f);
 result += GET_MATRIX_DATA(iD, Y+28,X+6, inWS) * (-0.00831752f);
 result += GET_MATRIX_DATA(iD, Y+30,X+6, inWS) * (-0.00588468f);
 result += GET_MATRIX_DATA(iD, Y+17,X+7, inWS) * (-0.02055914f);
 result += GET_MATRIX_DATA(iD, Y+6,X+8, inWS) * (-0.00689839f);
 result += GET_MATRIX_DATA(iD, Y+30,X+8, inWS) * (-0.00492028f);
 result += GET_MATRIX_DATA(iD, Y+12,X+11, inWS) * (0.01390506f);
 result += GET_MATRIX_DATA(iD, Y+14,X+11, inWS) * (0.00064087f);
 result += GET_MATRIX_DATA(iD, Y+15,X+11, inWS) * (0.00077979f);
 result += GET_MATRIX_DATA(iD, Y+20,X+11, inWS) * (0.00320967f);
 result += GET_MATRIX_DATA(iD, Y+21,X+11, inWS) * (0.03602699f);
 result += GET_MATRIX_DATA(iD, Y+12,X+12, inWS) * (0.01568222f);
 result += GET_MATRIX_DATA(iD, Y+13,X+12, inWS) * (0.00940344f);
 result += GET_MATRIX_DATA(iD, Y+22,X+12, inWS) * (0.01050997f);
 result += GET_MATRIX_DATA(iD, Y+11,X+13, inWS) * (0.05996817f);
 result += GET_MATRIX_DATA(iD, Y+13,X+13, inWS) * (0.01163602f);
 result += GET_MATRIX_DATA(iD, Y+22,X+13, inWS) * (0.06075170f);
 result += GET_MATRIX_DATA(iD, Y+23,X+13, inWS) * (0.04837013f);
 result += GET_MATRIX_DATA(iD, Y+24,X+13, inWS) * (0.04245766f);
 result += GET_MATRIX_DATA(iD, Y+11,X+14, inWS) * (0.05063013f);
 result += GET_MATRIX_DATA(iD, Y+23,X+14, inWS) * (0.00504614f);
 result += GET_MATRIX_DATA(iD, Y+16,X+15, inWS) * (-0.04133513f);
 result += GET_MATRIX_DATA(iD, Y+18,X+15, inWS) * (-0.07233403f);
 result += GET_MATRIX_DATA(iD, Y+23,X+15, inWS) * (0.00740993f);
 result += GET_MATRIX_DATA(iD, Y+8,X+17, inWS) * (-0.00351391f);
 result += GET_MATRIX_DATA(iD, Y+12,X+18, inWS) * (-0.02671837f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (-0.01052053f);
 result += GET_MATRIX_DATA(iD, Y+22,X+18, inWS) * (-0.02051612f);
 result += GET_MATRIX_DATA(iD, Y+26,X+18, inWS) * (-0.01652776f);
 result += GET_MATRIX_DATA(iD, Y+8,X+19, inWS) * (-0.00558798f);
 result += GET_MATRIX_DATA(iD, Y+17,X+19, inWS) * (0.09088852f);
 result += GET_MATRIX_DATA(iD, Y+20,X+19, inWS) * (0.00138338f);
 result += GET_MATRIX_DATA(iD, Y+24,X+19, inWS) * (-0.01038105f);
 result += GET_MATRIX_DATA(iD, Y+24,X+20, inWS) * (-0.02208806f);
 result += GET_MATRIX_DATA(iD, Y+17,X+22, inWS) * (0.01221614f);
 result += GET_MATRIX_DATA(iD, Y+4,X+24, inWS) * (0.00901978f);
 result += GET_MATRIX_DATA(iD, Y+5,X+24, inWS) * (0.00191409f);
 result += GET_MATRIX_DATA(iD, Y+16,X+24, inWS) * (0.01431862f);
 result += GET_MATRIX_DATA(iD, Y+30,X+24, inWS) * (0.00507579f);
 result += GET_MATRIX_DATA(iD, Y+9,X+25, inWS) * (0.00554044f);
 result += GET_MATRIX_DATA(iD, Y+10,X+25, inWS) * (0.01162508f);
 result += GET_MATRIX_DATA(iD, Y+30,X+25, inWS) * (0.00882410f);
 result += GET_MATRIX_DATA(iD, Y+3,X+26, inWS) * (0.00073871f);
 result += GET_MATRIX_DATA(iD, Y+7,X+26, inWS) * (0.00279457f);
 result += GET_MATRIX_DATA(iD, Y+28,X+26, inWS) * (0.00117461f);
 result += GET_MATRIX_DATA(iD, Y+30,X+26, inWS) * (0.00093875f);
 result += GET_MATRIX_DATA(iD, Y+31,X+26, inWS) * (0.01693134f);
 result += GET_MATRIX_DATA(iD, Y+30,X+28, inWS) * (0.01695715f);
 result += GET_MATRIX_DATA(iD, Y+14,X+30, inWS) * (-0.00093709f);
 result += GET_MATRIX_DATA(iD, Y+20,X+30, inWS) * (-0.00943804f);
 result += GET_MATRIX_DATA(iD, Y+17,X+31, inWS) * (-0.01829128f);
 result += GET_MATRIX_DATA(iD, Y+14,X+32, inWS) * (-0.00081684f);
 result += GET_MATRIX_DATA(iD, Y+15,X+32, inWS) * (-0.00637655f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00691724f);
desc[18] = (result > ( 7.35f));

 /***********************************************
                  Projection: 20
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+28,X+2, inWS) * (0.00256929f);
 result += GET_MATRIX_DATA(iD, Y+14,X+6, inWS) * (-0.00987737f);
 result += GET_MATRIX_DATA(iD, Y+18,X+6, inWS) * (-0.01834376f);
 result += GET_MATRIX_DATA(iD, Y+20,X+6, inWS) * (-0.00641017f);
 result += GET_MATRIX_DATA(iD, Y+21,X+7, inWS) * (-0.01093632f);
 result += GET_MATRIX_DATA(iD, Y+24,X+7, inWS) * (-0.01027218f);
 result += GET_MATRIX_DATA(iD, Y+26,X+9, inWS) * (-0.00624051f);
 result += GET_MATRIX_DATA(iD, Y+13,X+10, inWS) * (0.00883813f);
 result += GET_MATRIX_DATA(iD, Y+14,X+10, inWS) * (0.01908560f);
 result += GET_MATRIX_DATA(iD, Y+15,X+10, inWS) * (0.04130264f);
 result += GET_MATRIX_DATA(iD, Y+16,X+10, inWS) * (0.03325038f);
 result += GET_MATRIX_DATA(iD, Y+17,X+10, inWS) * (0.00146965f);
 result += GET_MATRIX_DATA(iD, Y+10,X+11, inWS) * (0.00860717f);
 result += GET_MATRIX_DATA(iD, Y+18,X+11, inWS) * (0.04235141f);
 result += GET_MATRIX_DATA(iD, Y+19,X+11, inWS) * (0.01953192f);
 result += GET_MATRIX_DATA(iD, Y+24,X+13, inWS) * (-0.00501561f);
 result += GET_MATRIX_DATA(iD, Y+25,X+13, inWS) * (-0.00642869f);
 result += GET_MATRIX_DATA(iD, Y+14,X+14, inWS) * (-0.05147215f);
 result += GET_MATRIX_DATA(iD, Y+15,X+14, inWS) * (-0.04666888f);
 result += GET_MATRIX_DATA(iD, Y+11,X+15, inWS) * (-0.00399896f);
 result += GET_MATRIX_DATA(iD, Y+14,X+15, inWS) * (-0.00747978f);
 result += GET_MATRIX_DATA(iD, Y+11,X+16, inWS) * (-0.06979138f);
 result += GET_MATRIX_DATA(iD, Y+24,X+16, inWS) * (-0.00716699f);
 result += GET_MATRIX_DATA(iD, Y+11,X+17, inWS) * (-0.05841617f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (0.01074166f);
 result += GET_MATRIX_DATA(iD, Y+17,X+17, inWS) * (0.00297441f);
 result += GET_MATRIX_DATA(iD, Y+5,X+18, inWS) * (0.00715660f);
 result += GET_MATRIX_DATA(iD, Y+10,X+18, inWS) * (-0.00368930f);
 result += GET_MATRIX_DATA(iD, Y+11,X+18, inWS) * (-0.00639216f);
 result += GET_MATRIX_DATA(iD, Y+14,X+18, inWS) * (0.09149558f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (0.06334603f);
 result += GET_MATRIX_DATA(iD, Y+16,X+18, inWS) * (0.03688449f);
 result += GET_MATRIX_DATA(iD, Y+14,X+19, inWS) * (0.00243867f);
 result += GET_MATRIX_DATA(iD, Y+18,X+20, inWS) * (-0.06549094f);
 result += GET_MATRIX_DATA(iD, Y+5,X+21, inWS) * (0.00222473f);
 result += GET_MATRIX_DATA(iD, Y+12,X+21, inWS) * (-0.00755213f);
 result += GET_MATRIX_DATA(iD, Y+16,X+21, inWS) * (-0.03576500f);
 result += GET_MATRIX_DATA(iD, Y+18,X+21, inWS) * (-0.00480601f);
 result += GET_MATRIX_DATA(iD, Y+19,X+21, inWS) * (-0.02128305f);
 result += GET_MATRIX_DATA(iD, Y+14,X+22, inWS) * (-0.00791254f);
 result += GET_MATRIX_DATA(iD, Y+15,X+22, inWS) * (-0.00635996f);
 result += GET_MATRIX_DATA(iD, Y+16,X+22, inWS) * (-0.04239763f);
 result += GET_MATRIX_DATA(iD, Y+17,X+22, inWS) * (-0.00982651f);
 result += GET_MATRIX_DATA(iD, Y+5,X+23, inWS) * (0.01309609f);
 result += GET_MATRIX_DATA(iD, Y+25,X+23, inWS) * (0.01293470f);
 result += GET_MATRIX_DATA(iD, Y+23,X+25, inWS) * (0.00171643f);
 result += GET_MATRIX_DATA(iD, Y+21,X+26, inWS) * (0.01046135f);
 result += GET_MATRIX_DATA(iD, Y+21,X+27, inWS) * (0.00037569f);
 result += GET_MATRIX_DATA(iD, Y+23,X+27, inWS) * (0.00202998f);
 result += GET_MATRIX_DATA(iD, Y+12,X+28, inWS) * (0.00556540f);
 result += GET_MATRIX_DATA(iD, Y+14,X+28, inWS) * (0.00823606f);
 result += GET_MATRIX_DATA(iD, Y+20,X+28, inWS) * (0.00001695f);
 result += GET_MATRIX_DATA(iD, Y+21,X+28, inWS) * (0.01203876f);
 result += GET_MATRIX_DATA(iD, Y+17,X+29, inWS) * (0.00701394f);
 result += GET_MATRIX_DATA(iD, Y+2,X+30, inWS) * (-0.00751856f);
 result += GET_MATRIX_DATA(iD, Y+3,X+30, inWS) * (-0.00340066f);
 result += GET_MATRIX_DATA(iD, Y+2,X+31, inWS) * (-0.00863871f);
 result += GET_MATRIX_DATA(iD, Y+3,X+31, inWS) * (-0.01213660f);
 result += GET_MATRIX_DATA(iD, Y+30,X+31, inWS) * (-0.01254871f);
 result += GET_MATRIX_DATA(iD, Y+32,X+31, inWS) * (-0.00183777f);
 result += GET_MATRIX_DATA(iD, Y+5,X+32, inWS) * (-0.01642849f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00380883f);
desc[19] = (result > (102.27f));

 /***********************************************
                  Projection: 21
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+7,X+1, inWS) * (0.00570550f);
 result += GET_MATRIX_DATA(iD, Y+9,X+2, inWS) * (0.01330777f);
 result += GET_MATRIX_DATA(iD, Y+12,X+2, inWS) * (0.00318152f);
 result += GET_MATRIX_DATA(iD, Y+6,X+3, inWS) * (0.01835534f);
 result += GET_MATRIX_DATA(iD, Y+10,X+3, inWS) * (0.01404212f);
 result += GET_MATRIX_DATA(iD, Y+11,X+3, inWS) * (0.01598773f);
 result += GET_MATRIX_DATA(iD, Y+12,X+3, inWS) * (0.01327088f);
 result += GET_MATRIX_DATA(iD, Y+22,X+3, inWS) * (-0.02392107f);
 result += GET_MATRIX_DATA(iD, Y+23,X+3, inWS) * (-0.00339582f);
 result += GET_MATRIX_DATA(iD, Y+24,X+3, inWS) * (-0.00344998f);
 result += GET_MATRIX_DATA(iD, Y+19,X+4, inWS) * (-0.00537900f);
 result += GET_MATRIX_DATA(iD, Y+21,X+5, inWS) * (-0.00154731f);
 result += GET_MATRIX_DATA(iD, Y+22,X+5, inWS) * (-0.01692456f);
 result += GET_MATRIX_DATA(iD, Y+21,X+6, inWS) * (-0.00002292f);
 result += GET_MATRIX_DATA(iD, Y+21,X+7, inWS) * (-0.00566940f);
 result += GET_MATRIX_DATA(iD, Y+2,X+9, inWS) * (-0.00365644f);
 result += GET_MATRIX_DATA(iD, Y+3,X+10, inWS) * (-0.03561851f);
 result += GET_MATRIX_DATA(iD, Y+30,X+11, inWS) * (0.02573556f);
 result += GET_MATRIX_DATA(iD, Y+32,X+11, inWS) * (0.00745569f);
 result += GET_MATRIX_DATA(iD, Y+1,X+12, inWS) * (-0.00273250f);
 result += GET_MATRIX_DATA(iD, Y+18,X+12, inWS) * (0.00583848f);
 result += GET_MATRIX_DATA(iD, Y+19,X+12, inWS) * (0.06237531f);
 result += GET_MATRIX_DATA(iD, Y+14,X+13, inWS) * (-0.00402083f);
 result += GET_MATRIX_DATA(iD, Y+15,X+13, inWS) * (-0.02021978f);
 result += GET_MATRIX_DATA(iD, Y+32,X+13, inWS) * (0.01500082f);
 result += GET_MATRIX_DATA(iD, Y+15,X+14, inWS) * (-0.01589246f);
 result += GET_MATRIX_DATA(iD, Y+17,X+15, inWS) * (0.02571488f);
 result += GET_MATRIX_DATA(iD, Y+18,X+15, inWS) * (0.02234782f);
 result += GET_MATRIX_DATA(iD, Y+8,X+16, inWS) * (0.01600459f);
 result += GET_MATRIX_DATA(iD, Y+10,X+16, inWS) * (0.01924394f);
 result += GET_MATRIX_DATA(iD, Y+11,X+16, inWS) * (0.02170702f);
 result += GET_MATRIX_DATA(iD, Y+15,X+16, inWS) * (-0.08389379f);
 result += GET_MATRIX_DATA(iD, Y+17,X+16, inWS) * (0.02700970f);
 result += GET_MATRIX_DATA(iD, Y+23,X+16, inWS) * (-0.05160003f);
 result += GET_MATRIX_DATA(iD, Y+24,X+16, inWS) * (-0.03092100f);
 result += GET_MATRIX_DATA(iD, Y+9,X+17, inWS) * (0.01037502f);
 result += GET_MATRIX_DATA(iD, Y+10,X+17, inWS) * (0.00764159f);
 result += GET_MATRIX_DATA(iD, Y+23,X+17, inWS) * (-0.06332393f);
 result += GET_MATRIX_DATA(iD, Y+24,X+17, inWS) * (-0.02492658f);
 result += GET_MATRIX_DATA(iD, Y+31,X+17, inWS) * (0.00339906f);
 result += GET_MATRIX_DATA(iD, Y+10,X+18, inWS) * (0.03404738f);
 result += GET_MATRIX_DATA(iD, Y+11,X+18, inWS) * (0.03811355f);
 result += GET_MATRIX_DATA(iD, Y+22,X+18, inWS) * (-0.02486268f);
 result += GET_MATRIX_DATA(iD, Y+31,X+19, inWS) * (0.00651465f);
 result += GET_MATRIX_DATA(iD, Y+27,X+21, inWS) * (0.00057731f);
 result += GET_MATRIX_DATA(iD, Y+30,X+21, inWS) * (0.01023588f);
 result += GET_MATRIX_DATA(iD, Y+4,X+22, inWS) * (-0.00399333f);
 result += GET_MATRIX_DATA(iD, Y+19,X+22, inWS) * (0.03905941f);
 result += GET_MATRIX_DATA(iD, Y+20,X+22, inWS) * (0.00810385f);
 result += GET_MATRIX_DATA(iD, Y+6,X+23, inWS) * (-0.02062598f);
 result += GET_MATRIX_DATA(iD, Y+13,X+23, inWS) * (-0.01417844f);
 result += GET_MATRIX_DATA(iD, Y+20,X+23, inWS) * (0.00381445f);
 result += GET_MATRIX_DATA(iD, Y+21,X+23, inWS) * (0.01081688f);
 result += GET_MATRIX_DATA(iD, Y+26,X+23, inWS) * (0.00786693f);
 result += GET_MATRIX_DATA(iD, Y+28,X+23, inWS) * (0.00753953f);
 result += GET_MATRIX_DATA(iD, Y+10,X+24, inWS) * (-0.00220681f);
 result += GET_MATRIX_DATA(iD, Y+13,X+24, inWS) * (-0.01075362f);
 result += GET_MATRIX_DATA(iD, Y+21,X+24, inWS) * (0.01101923f);
 result += GET_MATRIX_DATA(iD, Y+6,X+30, inWS) * (0.00002785f);
 result += GET_MATRIX_DATA(iD, Y+26,X+30, inWS) * (-0.00883442f);
 result += GET_MATRIX_DATA(iD, Y+28,X+30, inWS) * (-0.00249058f);
 result += GET_MATRIX_DATA(iD, Y+6,X+31, inWS) * (0.00615275f);
 result += GET_MATRIX_DATA(iD, Y+30,X+31, inWS) * (-0.01584618f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00041630f);
desc[20] = (result > (17.20f));

 /***********************************************
                  Projection: 22
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+21,X+3, inWS) * (0.00728627f);
 result += GET_MATRIX_DATA(iD, Y+1,X+5, inWS) * (0.02011960f);
 result += GET_MATRIX_DATA(iD, Y+12,X+5, inWS) * (-0.00480183f);
 result += GET_MATRIX_DATA(iD, Y+23,X+5, inWS) * (0.00896111f);
 result += GET_MATRIX_DATA(iD, Y+10,X+6, inWS) * (-0.00499789f);
 result += GET_MATRIX_DATA(iD, Y+13,X+6, inWS) * (-0.00278804f);
 result += GET_MATRIX_DATA(iD, Y+21,X+6, inWS) * (0.01683119f);
 result += GET_MATRIX_DATA(iD, Y+22,X+6, inWS) * (0.00138824f);
 result += GET_MATRIX_DATA(iD, Y+8,X+7, inWS) * (-0.00589855f);
 result += GET_MATRIX_DATA(iD, Y+10,X+7, inWS) * (-0.00938516f);
 result += GET_MATRIX_DATA(iD, Y+24,X+7, inWS) * (0.00290480f);
 result += GET_MATRIX_DATA(iD, Y+3,X+8, inWS) * (0.00806885f);
 result += GET_MATRIX_DATA(iD, Y+12,X+8, inWS) * (-0.02163526f);
 result += GET_MATRIX_DATA(iD, Y+20,X+8, inWS) * (0.00532340f);
 result += GET_MATRIX_DATA(iD, Y+10,X+9, inWS) * (-0.00660485f);
 result += GET_MATRIX_DATA(iD, Y+2,X+10, inWS) * (0.00707226f);
 result += GET_MATRIX_DATA(iD, Y+3,X+10, inWS) * (0.00921139f);
 result += GET_MATRIX_DATA(iD, Y+10,X+10, inWS) * (-0.00956705f);
 result += GET_MATRIX_DATA(iD, Y+2,X+12, inWS) * (0.01233607f);
 result += GET_MATRIX_DATA(iD, Y+3,X+12, inWS) * (0.00650221f);
 result += GET_MATRIX_DATA(iD, Y+15,X+13, inWS) * (0.02322230f);
 result += GET_MATRIX_DATA(iD, Y+28,X+13, inWS) * (-0.00636845f);
 result += GET_MATRIX_DATA(iD, Y+25,X+14, inWS) * (-0.01189005f);
 result += GET_MATRIX_DATA(iD, Y+23,X+16, inWS) * (-0.00658486f);
 result += GET_MATRIX_DATA(iD, Y+9,X+18, inWS) * (0.00151103f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (0.05103108f);
 result += GET_MATRIX_DATA(iD, Y+30,X+18, inWS) * (0.00280735f);
 result += GET_MATRIX_DATA(iD, Y+31,X+18, inWS) * (0.00013887f);
 result += GET_MATRIX_DATA(iD, Y+19,X+19, inWS) * (0.01373360f);
 result += GET_MATRIX_DATA(iD, Y+30,X+19, inWS) * (0.00603175f);
 result += GET_MATRIX_DATA(iD, Y+22,X+20, inWS) * (-0.05125539f);
 result += GET_MATRIX_DATA(iD, Y+28,X+20, inWS) * (0.00857769f);
 result += GET_MATRIX_DATA(iD, Y+30,X+20, inWS) * (0.03744423f);
 result += GET_MATRIX_DATA(iD, Y+31,X+20, inWS) * (0.00583269f);
 result += GET_MATRIX_DATA(iD, Y+3,X+21, inWS) * (-0.01877864f);
 result += GET_MATRIX_DATA(iD, Y+21,X+21, inWS) * (-0.00379702f);
 result += GET_MATRIX_DATA(iD, Y+2,X+22, inWS) * (-0.01386991f);
 result += GET_MATRIX_DATA(iD, Y+3,X+22, inWS) * (-0.00183379f);
 result += GET_MATRIX_DATA(iD, Y+10,X+22, inWS) * (0.00227806f);
 result += GET_MATRIX_DATA(iD, Y+15,X+22, inWS) * (-0.01298300f);
 result += GET_MATRIX_DATA(iD, Y+28,X+22, inWS) * (0.01151126f);
 result += GET_MATRIX_DATA(iD, Y+30,X+22, inWS) * (0.02455191f);
 result += GET_MATRIX_DATA(iD, Y+32,X+22, inWS) * (0.00045688f);
 result += GET_MATRIX_DATA(iD, Y+3,X+23, inWS) * (-0.02736030f);
 result += GET_MATRIX_DATA(iD, Y+18,X+23, inWS) * (-0.00558801f);
 result += GET_MATRIX_DATA(iD, Y+19,X+23, inWS) * (-0.00703338f);
 result += GET_MATRIX_DATA(iD, Y+20,X+23, inWS) * (-0.02460552f);
 result += GET_MATRIX_DATA(iD, Y+30,X+23, inWS) * (0.00802797f);
 result += GET_MATRIX_DATA(iD, Y+30,X+24, inWS) * (0.02750200f);
 result += GET_MATRIX_DATA(iD, Y+11,X+26, inWS) * (0.01657524f);
 result += GET_MATRIX_DATA(iD, Y+10,X+27, inWS) * (0.01355435f);
 result += GET_MATRIX_DATA(iD, Y+14,X+27, inWS) * (0.00177555f);
 result += GET_MATRIX_DATA(iD, Y+12,X+28, inWS) * (0.02312519f);
 result += GET_MATRIX_DATA(iD, Y+13,X+28, inWS) * (0.00615353f);
 result += GET_MATRIX_DATA(iD, Y+10,X+30, inWS) * (0.00353504f);
 result += GET_MATRIX_DATA(iD, Y+12,X+30, inWS) * (0.02168259f);
 result += GET_MATRIX_DATA(iD, Y+13,X+30, inWS) * (0.00021633f);
 result += GET_MATRIX_DATA(iD, Y+23,X+30, inWS) * (-0.01192229f);
 result += GET_MATRIX_DATA(iD, Y+25,X+30, inWS) * (-0.00039194f);
 result += GET_MATRIX_DATA(iD, Y+22,X+31, inWS) * (-0.00617866f);
 result += GET_MATRIX_DATA(iD, Y+27,X+32, inWS) * (-0.01301342f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00217889f);
desc[21] = (result > (106.87f));

 /***********************************************
                  Projection: 23
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+17,X+1, inWS) * (-0.00023719f);
 result += GET_MATRIX_DATA(iD, Y+10,X+6, inWS) * (0.00098793f);
 result += GET_MATRIX_DATA(iD, Y+11,X+6, inWS) * (0.00564689f);
 result += GET_MATRIX_DATA(iD, Y+9,X+7, inWS) * (0.01330257f);
 result += GET_MATRIX_DATA(iD, Y+26,X+7, inWS) * (0.00141865f);
 result += GET_MATRIX_DATA(iD, Y+8,X+8, inWS) * (0.00271295f);
 result += GET_MATRIX_DATA(iD, Y+25,X+8, inWS) * (0.00550810f);
 result += GET_MATRIX_DATA(iD, Y+27,X+8, inWS) * (0.00194239f);
 result += GET_MATRIX_DATA(iD, Y+8,X+9, inWS) * (0.00420299f);
 result += GET_MATRIX_DATA(iD, Y+16,X+9, inWS) * (-0.02117054f);
 result += GET_MATRIX_DATA(iD, Y+17,X+9, inWS) * (-0.00398405f);
 result += GET_MATRIX_DATA(iD, Y+25,X+9, inWS) * (0.00517948f);
 result += GET_MATRIX_DATA(iD, Y+17,X+10, inWS) * (-0.00319066f);
 result += GET_MATRIX_DATA(iD, Y+18,X+10, inWS) * (-0.04203405f);
 result += GET_MATRIX_DATA(iD, Y+19,X+10, inWS) * (-0.01048592f);
 result += GET_MATRIX_DATA(iD, Y+26,X+10, inWS) * (0.00729924f);
 result += GET_MATRIX_DATA(iD, Y+8,X+11, inWS) * (0.01497456f);
 result += GET_MATRIX_DATA(iD, Y+15,X+11, inWS) * (-0.00521381f);
 result += GET_MATRIX_DATA(iD, Y+16,X+11, inWS) * (-0.01227685f);
 result += GET_MATRIX_DATA(iD, Y+17,X+11, inWS) * (-0.03440689f);
 result += GET_MATRIX_DATA(iD, Y+25,X+12, inWS) * (0.01572968f);
 result += GET_MATRIX_DATA(iD, Y+1,X+13, inWS) * (-0.00196513f);
 result += GET_MATRIX_DATA(iD, Y+25,X+13, inWS) * (0.00012878f);
 result += GET_MATRIX_DATA(iD, Y+1,X+14, inWS) * (-0.03130822f);
 result += GET_MATRIX_DATA(iD, Y+3,X+14, inWS) * (-0.00538353f);
 result += GET_MATRIX_DATA(iD, Y+18,X+14, inWS) * (0.00403744f);
 result += GET_MATRIX_DATA(iD, Y+26,X+14, inWS) * (0.00131828f);
 result += GET_MATRIX_DATA(iD, Y+3,X+15, inWS) * (-0.02996797f);
 result += GET_MATRIX_DATA(iD, Y+18,X+16, inWS) * (0.03325016f);
 result += GET_MATRIX_DATA(iD, Y+30,X+16, inWS) * (-0.00098747f);
 result += GET_MATRIX_DATA(iD, Y+32,X+16, inWS) * (-0.02873084f);
 result += GET_MATRIX_DATA(iD, Y+1,X+17, inWS) * (-0.01244706f);
 result += GET_MATRIX_DATA(iD, Y+23,X+17, inWS) * (-0.00914790f);
 result += GET_MATRIX_DATA(iD, Y+1,X+18, inWS) * (-0.02411301f);
 result += GET_MATRIX_DATA(iD, Y+11,X+18, inWS) * (-0.01851480f);
 result += GET_MATRIX_DATA(iD, Y+20,X+18, inWS) * (-0.02859380f);
 result += GET_MATRIX_DATA(iD, Y+21,X+18, inWS) * (-0.00737122f);
 result += GET_MATRIX_DATA(iD, Y+23,X+18, inWS) * (-0.00506239f);
 result += GET_MATRIX_DATA(iD, Y+30,X+18, inWS) * (-0.01502231f);
 result += GET_MATRIX_DATA(iD, Y+16,X+19, inWS) * (0.04550980f);
 result += GET_MATRIX_DATA(iD, Y+25,X+21, inWS) * (0.00013971f);
 result += GET_MATRIX_DATA(iD, Y+7,X+22, inWS) * (0.01276889f);
 result += GET_MATRIX_DATA(iD, Y+14,X+22, inWS) * (-0.04377926f);
 result += GET_MATRIX_DATA(iD, Y+24,X+22, inWS) * (0.00540753f);
 result += GET_MATRIX_DATA(iD, Y+25,X+22, inWS) * (0.00679793f);
 result += GET_MATRIX_DATA(iD, Y+9,X+23, inWS) * (0.00686215f);
 result += GET_MATRIX_DATA(iD, Y+16,X+23, inWS) * (-0.03552178f);
 result += GET_MATRIX_DATA(iD, Y+17,X+23, inWS) * (-0.00080884f);
 result += GET_MATRIX_DATA(iD, Y+18,X+23, inWS) * (-0.00910550f);
 result += GET_MATRIX_DATA(iD, Y+24,X+23, inWS) * (0.01339924f);
 result += GET_MATRIX_DATA(iD, Y+26,X+23, inWS) * (0.02709725f);
 result += GET_MATRIX_DATA(iD, Y+7,X+24, inWS) * (0.03306426f);
 result += GET_MATRIX_DATA(iD, Y+17,X+24, inWS) * (-0.00357160f);
 result += GET_MATRIX_DATA(iD, Y+18,X+24, inWS) * (-0.00542055f);
 result += GET_MATRIX_DATA(iD, Y+25,X+24, inWS) * (0.00648488f);
 result += GET_MATRIX_DATA(iD, Y+27,X+24, inWS) * (0.00557402f);
 result += GET_MATRIX_DATA(iD, Y+9,X+25, inWS) * (0.03398196f);
 result += GET_MATRIX_DATA(iD, Y+25,X+25, inWS) * (0.00547501f);
 result += GET_MATRIX_DATA(iD, Y+6,X+26, inWS) * (0.00028870f);
 result += GET_MATRIX_DATA(iD, Y+8,X+27, inWS) * (0.00116124f);
 result += GET_MATRIX_DATA(iD, Y+31,X+31, inWS) * (-0.00254617f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00226036f);
desc[22] = (result > (43.27f));

 /***********************************************
                  Projection: 24
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (-0.01043807f);
 result += GET_MATRIX_DATA(iD, Y+31,X+3, inWS) * (-0.01959241f);
 result += GET_MATRIX_DATA(iD, Y+12,X+5, inWS) * (-0.00270220f);
 result += GET_MATRIX_DATA(iD, Y+13,X+6, inWS) * (-0.01345012f);
 result += GET_MATRIX_DATA(iD, Y+30,X+6, inWS) * (-0.00399754f);
 result += GET_MATRIX_DATA(iD, Y+32,X+6, inWS) * (-0.00770206f);
 result += GET_MATRIX_DATA(iD, Y+12,X+7, inWS) * (-0.00237197f);
 result += GET_MATRIX_DATA(iD, Y+12,X+8, inWS) * (-0.00719104f);
 result += GET_MATRIX_DATA(iD, Y+11,X+9, inWS) * (-0.01630489f);
 result += GET_MATRIX_DATA(iD, Y+25,X+9, inWS) * (0.01354440f);
 result += GET_MATRIX_DATA(iD, Y+17,X+10, inWS) * (0.01220154f);
 result += GET_MATRIX_DATA(iD, Y+18,X+10, inWS) * (0.01737336f);
 result += GET_MATRIX_DATA(iD, Y+25,X+10, inWS) * (0.00377160f);
 result += GET_MATRIX_DATA(iD, Y+26,X+10, inWS) * (0.00314621f);
 result += GET_MATRIX_DATA(iD, Y+3,X+11, inWS) * (0.00047754f);
 result += GET_MATRIX_DATA(iD, Y+5,X+12, inWS) * (0.00332634f);
 result += GET_MATRIX_DATA(iD, Y+12,X+12, inWS) * (-0.01344105f);
 result += GET_MATRIX_DATA(iD, Y+25,X+12, inWS) * (0.00651862f);
 result += GET_MATRIX_DATA(iD, Y+26,X+12, inWS) * (0.00775259f);
 result += GET_MATRIX_DATA(iD, Y+5,X+13, inWS) * (0.00435298f);
 result += GET_MATRIX_DATA(iD, Y+13,X+14, inWS) * (-0.03096035f);
 result += GET_MATRIX_DATA(iD, Y+14,X+14, inWS) * (-0.04546240f);
 result += GET_MATRIX_DATA(iD, Y+20,X+14, inWS) * (-0.01627842f);
 result += GET_MATRIX_DATA(iD, Y+21,X+14, inWS) * (-0.04259330f);
 result += GET_MATRIX_DATA(iD, Y+22,X+14, inWS) * (-0.00786220f);
 result += GET_MATRIX_DATA(iD, Y+22,X+15, inWS) * (-0.04062218f);
 result += GET_MATRIX_DATA(iD, Y+16,X+16, inWS) * (0.02779079f);
 result += GET_MATRIX_DATA(iD, Y+17,X+16, inWS) * (0.10235525f);
 result += GET_MATRIX_DATA(iD, Y+18,X+16, inWS) * (0.00807545f);
 result += GET_MATRIX_DATA(iD, Y+22,X+16, inWS) * (-0.00621445f);
 result += GET_MATRIX_DATA(iD, Y+3,X+18, inWS) * (-0.00220892f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (-0.03582905f);
 result += GET_MATRIX_DATA(iD, Y+16,X+18, inWS) * (-0.10260119f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (-0.03855744f);
 result += GET_MATRIX_DATA(iD, Y+30,X+18, inWS) * (-0.02908384f);
 result += GET_MATRIX_DATA(iD, Y+3,X+19, inWS) * (-0.02174429f);
 result += GET_MATRIX_DATA(iD, Y+4,X+19, inWS) * (-0.00921033f);
 result += GET_MATRIX_DATA(iD, Y+13,X+19, inWS) * (0.07687172f);
 result += GET_MATRIX_DATA(iD, Y+20,X+19, inWS) * (0.05127618f);
 result += GET_MATRIX_DATA(iD, Y+13,X+20, inWS) * (0.02397417f);
 result += GET_MATRIX_DATA(iD, Y+21,X+20, inWS) * (0.04410605f);
 result += GET_MATRIX_DATA(iD, Y+3,X+21, inWS) * (-0.01656841f);
 result += GET_MATRIX_DATA(iD, Y+12,X+21, inWS) * (0.02550962f);
 result += GET_MATRIX_DATA(iD, Y+16,X+21, inWS) * (-0.02555932f);
 result += GET_MATRIX_DATA(iD, Y+22,X+21, inWS) * (0.00664165f);
 result += GET_MATRIX_DATA(iD, Y+13,X+22, inWS) * (0.00577677f);
 result += GET_MATRIX_DATA(iD, Y+14,X+22, inWS) * (0.01770863f);
 result += GET_MATRIX_DATA(iD, Y+11,X+23, inWS) * (0.01124271f);
 result += GET_MATRIX_DATA(iD, Y+12,X+23, inWS) * (0.00265226f);
 result += GET_MATRIX_DATA(iD, Y+17,X+27, inWS) * (-0.02701305f);
 result += GET_MATRIX_DATA(iD, Y+18,X+27, inWS) * (-0.01660444f);
 result += GET_MATRIX_DATA(iD, Y+20,X+27, inWS) * (-0.01158936f);
 result += GET_MATRIX_DATA(iD, Y+21,X+27, inWS) * (-0.00205372f);
 result += GET_MATRIX_DATA(iD, Y+30,X+27, inWS) * (0.00157589f);
 result += GET_MATRIX_DATA(iD, Y+16,X+28, inWS) * (-0.00868891f);
 result += GET_MATRIX_DATA(iD, Y+17,X+28, inWS) * (-0.00503582f);
 result += GET_MATRIX_DATA(iD, Y+19,X+28, inWS) * (-0.00311057f);
 result += GET_MATRIX_DATA(iD, Y+29,X+29, inWS) * (0.00151317f);
 result += GET_MATRIX_DATA(iD, Y+16,X+30, inWS) * (-0.00839735f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00380739f);
desc[23] = (result > (-0.14f));

 /***********************************************
                  Projection: 25
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+24,X+3, inWS) * (0.00745247f);
 result += GET_MATRIX_DATA(iD, Y+24,X+4, inWS) * (0.00265813f);
 result += GET_MATRIX_DATA(iD, Y+17,X+7, inWS) * (-0.00957898f);
 result += GET_MATRIX_DATA(iD, Y+18,X+7, inWS) * (-0.03457883f);
 result += GET_MATRIX_DATA(iD, Y+16,X+8, inWS) * (-0.03291453f);
 result += GET_MATRIX_DATA(iD, Y+18,X+8, inWS) * (-0.00087032f);
 result += GET_MATRIX_DATA(iD, Y+20,X+11, inWS) * (0.00508820f);
 result += GET_MATRIX_DATA(iD, Y+20,X+12, inWS) * (0.01974189f);
 result += GET_MATRIX_DATA(iD, Y+21,X+12, inWS) * (0.00069390f);
 result += GET_MATRIX_DATA(iD, Y+13,X+13, inWS) * (0.00791210f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (-0.02039498f);
 result += GET_MATRIX_DATA(iD, Y+18,X+13, inWS) * (-0.04602175f);
 result += GET_MATRIX_DATA(iD, Y+21,X+13, inWS) * (0.06146197f);
 result += GET_MATRIX_DATA(iD, Y+4,X+14, inWS) * (0.00027039f);
 result += GET_MATRIX_DATA(iD, Y+13,X+14, inWS) * (0.03359609f);
 result += GET_MATRIX_DATA(iD, Y+14,X+14, inWS) * (0.04463794f);
 result += GET_MATRIX_DATA(iD, Y+21,X+14, inWS) * (0.04076382f);
 result += GET_MATRIX_DATA(iD, Y+14,X+15, inWS) * (0.03956725f);
 result += GET_MATRIX_DATA(iD, Y+8,X+16, inWS) * (0.00673403f);
 result += GET_MATRIX_DATA(iD, Y+11,X+16, inWS) * (-0.07343096f);
 result += GET_MATRIX_DATA(iD, Y+11,X+17, inWS) * (-0.05447454f);
 result += GET_MATRIX_DATA(iD, Y+12,X+17, inWS) * (-0.00010724f);
 result += GET_MATRIX_DATA(iD, Y+20,X+17, inWS) * (-0.04189093f);
 result += GET_MATRIX_DATA(iD, Y+22,X+17, inWS) * (-0.01870871f);
 result += GET_MATRIX_DATA(iD, Y+13,X+18, inWS) * (-0.00823469f);
 result += GET_MATRIX_DATA(iD, Y+14,X+18, inWS) * (-0.02639885f);
 result += GET_MATRIX_DATA(iD, Y+22,X+18, inWS) * (-0.01843495f);
 result += GET_MATRIX_DATA(iD, Y+1,X+19, inWS) * (-0.00166956f);
 result += GET_MATRIX_DATA(iD, Y+8,X+19, inWS) * (0.01582707f);
 result += GET_MATRIX_DATA(iD, Y+10,X+20, inWS) * (0.04911159f);
 result += GET_MATRIX_DATA(iD, Y+20,X+20, inWS) * (-0.01465982f);
 result += GET_MATRIX_DATA(iD, Y+24,X+20, inWS) * (0.01759338f);
 result += GET_MATRIX_DATA(iD, Y+11,X+21, inWS) * (0.02366586f);
 result += GET_MATRIX_DATA(iD, Y+12,X+21, inWS) * (0.00451733f);
 result += GET_MATRIX_DATA(iD, Y+25,X+21, inWS) * (0.01550219f);
 result += GET_MATRIX_DATA(iD, Y+13,X+22, inWS) * (0.05130378f);
 result += GET_MATRIX_DATA(iD, Y+14,X+22, inWS) * (0.03331305f);
 result += GET_MATRIX_DATA(iD, Y+15,X+22, inWS) * (0.02082374f);
 result += GET_MATRIX_DATA(iD, Y+23,X+22, inWS) * (0.01148264f);
 result += GET_MATRIX_DATA(iD, Y+32,X+22, inWS) * (-0.00446808f);
 result += GET_MATRIX_DATA(iD, Y+1,X+23, inWS) * (-0.02314300f);
 result += GET_MATRIX_DATA(iD, Y+3,X+23, inWS) * (-0.00260981f);
 result += GET_MATRIX_DATA(iD, Y+3,X+25, inWS) * (-0.01089551f);
 result += GET_MATRIX_DATA(iD, Y+4,X+25, inWS) * (-0.00207879f);
 result += GET_MATRIX_DATA(iD, Y+6,X+25, inWS) * (-0.00179504f);
 result += GET_MATRIX_DATA(iD, Y+12,X+25, inWS) * (-0.00372274f);
 result += GET_MATRIX_DATA(iD, Y+13,X+25, inWS) * (-0.01643968f);
 result += GET_MATRIX_DATA(iD, Y+10,X+26, inWS) * (-0.00131668f);
 result += GET_MATRIX_DATA(iD, Y+13,X+26, inWS) * (-0.00872651f);
 result += GET_MATRIX_DATA(iD, Y+14,X+26, inWS) * (-0.00758303f);
 result += GET_MATRIX_DATA(iD, Y+16,X+26, inWS) * (-0.00080080f);
 result += GET_MATRIX_DATA(iD, Y+17,X+26, inWS) * (-0.00049304f);
 result += GET_MATRIX_DATA(iD, Y+18,X+26, inWS) * (-0.03349592f);
 result += GET_MATRIX_DATA(iD, Y+20,X+26, inWS) * (-0.00605871f);
 result += GET_MATRIX_DATA(iD, Y+9,X+31, inWS) * (0.00129256f);
 result += GET_MATRIX_DATA(iD, Y+10,X+31, inWS) * (0.00773970f);
 result += GET_MATRIX_DATA(iD, Y+5,X+32, inWS) * (0.00365305f);
 result += GET_MATRIX_DATA(iD, Y+8,X+32, inWS) * (0.00210750f);
 result += GET_MATRIX_DATA(iD, Y+9,X+32, inWS) * (0.00241019f);
 result += GET_MATRIX_DATA(iD, Y+11,X+32, inWS) * (0.00223256f);
 result += GET_MATRIX_DATA(iD, Y+12,X+32, inWS) * (0.00054058f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00003063f);
desc[24] = (result > (54.74f));

 /***********************************************
                  Projection: 26
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+16,X+2, inWS) * (0.00730631f);
 result += GET_MATRIX_DATA(iD, Y+26,X+2, inWS) * (-0.00166836f);
 result += GET_MATRIX_DATA(iD, Y+28,X+2, inWS) * (-0.00071682f);
 result += GET_MATRIX_DATA(iD, Y+26,X+3, inWS) * (-0.00854956f);
 result += GET_MATRIX_DATA(iD, Y+15,X+4, inWS) * (0.00514711f);
 result += GET_MATRIX_DATA(iD, Y+16,X+4, inWS) * (0.00025054f);
 result += GET_MATRIX_DATA(iD, Y+16,X+5, inWS) * (0.01210328f);
 result += GET_MATRIX_DATA(iD, Y+17,X+5, inWS) * (0.03116393f);
 result += GET_MATRIX_DATA(iD, Y+27,X+5, inWS) * (-0.00085246f);
 result += GET_MATRIX_DATA(iD, Y+19,X+6, inWS) * (0.01418709f);
 result += GET_MATRIX_DATA(iD, Y+8,X+8, inWS) * (-0.00525383f);
 result += GET_MATRIX_DATA(iD, Y+32,X+9, inWS) * (0.00108511f);
 result += GET_MATRIX_DATA(iD, Y+1,X+10, inWS) * (0.02004721f);
 result += GET_MATRIX_DATA(iD, Y+3,X+10, inWS) * (0.00228086f);
 result += GET_MATRIX_DATA(iD, Y+8,X+10, inWS) * (-0.01396541f);
 result += GET_MATRIX_DATA(iD, Y+9,X+10, inWS) * (-0.00492119f);
 result += GET_MATRIX_DATA(iD, Y+10,X+10, inWS) * (-0.00188020f);
 result += GET_MATRIX_DATA(iD, Y+25,X+10, inWS) * (-0.02659838f);
 result += GET_MATRIX_DATA(iD, Y+32,X+10, inWS) * (0.00678503f);
 result += GET_MATRIX_DATA(iD, Y+3,X+11, inWS) * (0.00762125f);
 result += GET_MATRIX_DATA(iD, Y+8,X+11, inWS) * (-0.00443679f);
 result += GET_MATRIX_DATA(iD, Y+32,X+11, inWS) * (0.01406829f);
 result += GET_MATRIX_DATA(iD, Y+8,X+12, inWS) * (-0.02158894f);
 result += GET_MATRIX_DATA(iD, Y+24,X+12, inWS) * (-0.00543211f);
 result += GET_MATRIX_DATA(iD, Y+25,X+12, inWS) * (-0.00135884f);
 result += GET_MATRIX_DATA(iD, Y+32,X+12, inWS) * (0.01759029f);
 result += GET_MATRIX_DATA(iD, Y+1,X+13, inWS) * (0.02491499f);
 result += GET_MATRIX_DATA(iD, Y+3,X+13, inWS) * (0.00198899f);
 result += GET_MATRIX_DATA(iD, Y+12,X+13, inWS) * (0.00806909f);
 result += GET_MATRIX_DATA(iD, Y+17,X+13, inWS) * (-0.00261999f);
 result += GET_MATRIX_DATA(iD, Y+20,X+13, inWS) * (0.06356539f);
 result += GET_MATRIX_DATA(iD, Y+25,X+13, inWS) * (-0.00107639f);
 result += GET_MATRIX_DATA(iD, Y+31,X+13, inWS) * (0.03353426f);
 result += GET_MATRIX_DATA(iD, Y+1,X+14, inWS) * (0.00560075f);
 result += GET_MATRIX_DATA(iD, Y+2,X+14, inWS) * (0.01683095f);
 result += GET_MATRIX_DATA(iD, Y+8,X+14, inWS) * (-0.01124088f);
 result += GET_MATRIX_DATA(iD, Y+12,X+14, inWS) * (0.00330729f);
 result += GET_MATRIX_DATA(iD, Y+17,X+14, inWS) * (-0.03449845f);
 result += GET_MATRIX_DATA(iD, Y+20,X+14, inWS) * (0.04019851f);
 result += GET_MATRIX_DATA(iD, Y+25,X+14, inWS) * (-0.01650823f);
 result += GET_MATRIX_DATA(iD, Y+13,X+16, inWS) * (0.03359757f);
 result += GET_MATRIX_DATA(iD, Y+25,X+16, inWS) * (-0.00782997f);
 result += GET_MATRIX_DATA(iD, Y+27,X+17, inWS) * (-0.00458638f);
 result += GET_MATRIX_DATA(iD, Y+19,X+18, inWS) * (-0.04393249f);
 result += GET_MATRIX_DATA(iD, Y+27,X+18, inWS) * (-0.02674169f);
 result += GET_MATRIX_DATA(iD, Y+22,X+20, inWS) * (0.00599298f);
 result += GET_MATRIX_DATA(iD, Y+20,X+21, inWS) * (0.02687076f);
 result += GET_MATRIX_DATA(iD, Y+21,X+21, inWS) * (0.04605632f);
 result += GET_MATRIX_DATA(iD, Y+22,X+21, inWS) * (0.02930219f);
 result += GET_MATRIX_DATA(iD, Y+15,X+22, inWS) * (-0.03306958f);
 result += GET_MATRIX_DATA(iD, Y+16,X+22, inWS) * (-0.00916754f);
 result += GET_MATRIX_DATA(iD, Y+8,X+23, inWS) * (0.01450458f);
 result += GET_MATRIX_DATA(iD, Y+8,X+25, inWS) * (0.00528500f);
 result += GET_MATRIX_DATA(iD, Y+9,X+25, inWS) * (0.01606075f);
 result += GET_MATRIX_DATA(iD, Y+11,X+25, inWS) * (0.00922732f);
 result += GET_MATRIX_DATA(iD, Y+23,X+25, inWS) * (-0.00266894f);
 result += GET_MATRIX_DATA(iD, Y+10,X+27, inWS) * (0.00114128f);
 result += GET_MATRIX_DATA(iD, Y+10,X+28, inWS) * (0.00435635f);
 result += GET_MATRIX_DATA(iD, Y+12,X+28, inWS) * (0.00378496f);
 result += GET_MATRIX_DATA(iD, Y+1,X+30, inWS) * (-0.00230372f);
 result += GET_MATRIX_DATA(iD, Y+2,X+30, inWS) * (-0.01579414f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00426392f);
desc[25] = (result > (13.01f));

 /***********************************************
                  Projection: 27
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+14,X+3, inWS) * (0.00139310f);
 result += GET_MATRIX_DATA(iD, Y+16,X+3, inWS) * (0.00870133f);
 result += GET_MATRIX_DATA(iD, Y+16,X+5, inWS) * (0.01069069f);
 result += GET_MATRIX_DATA(iD, Y+16,X+6, inWS) * (0.00531785f);
 result += GET_MATRIX_DATA(iD, Y+15,X+7, inWS) * (0.01230197f);
 result += GET_MATRIX_DATA(iD, Y+16,X+7, inWS) * (0.01311927f);
 result += GET_MATRIX_DATA(iD, Y+24,X+7, inWS) * (-0.00025060f);
 result += GET_MATRIX_DATA(iD, Y+16,X+8, inWS) * (0.01709036f);
 result += GET_MATRIX_DATA(iD, Y+17,X+8, inWS) * (0.00825682f);
 result += GET_MATRIX_DATA(iD, Y+17,X+9, inWS) * (0.01038516f);
 result += GET_MATRIX_DATA(iD, Y+31,X+10, inWS) * (0.00043912f);
 result += GET_MATRIX_DATA(iD, Y+32,X+10, inWS) * (0.00928441f);
 result += GET_MATRIX_DATA(iD, Y+32,X+11, inWS) * (0.00456470f);
 result += GET_MATRIX_DATA(iD, Y+10,X+12, inWS) * (-0.00951697f);
 result += GET_MATRIX_DATA(iD, Y+11,X+12, inWS) * (-0.01360766f);
 result += GET_MATRIX_DATA(iD, Y+32,X+12, inWS) * (0.01695473f);
 result += GET_MATRIX_DATA(iD, Y+11,X+13, inWS) * (-0.00606614f);
 result += GET_MATRIX_DATA(iD, Y+15,X+13, inWS) * (0.02867091f);
 result += GET_MATRIX_DATA(iD, Y+31,X+13, inWS) * (0.00321336f);
 result += GET_MATRIX_DATA(iD, Y+11,X+14, inWS) * (-0.02329919f);
 result += GET_MATRIX_DATA(iD, Y+18,X+14, inWS) * (-0.01034997f);
 result += GET_MATRIX_DATA(iD, Y+19,X+14, inWS) * (-0.01474782f);
 result += GET_MATRIX_DATA(iD, Y+17,X+15, inWS) * (-0.01438695f);
 result += GET_MATRIX_DATA(iD, Y+18,X+15, inWS) * (-0.00353245f);
 result += GET_MATRIX_DATA(iD, Y+22,X+15, inWS) * (0.02366833f);
 result += GET_MATRIX_DATA(iD, Y+22,X+16, inWS) * (0.07014308f);
 result += GET_MATRIX_DATA(iD, Y+26,X+16, inWS) * (-0.02243238f);
 result += GET_MATRIX_DATA(iD, Y+19,X+17, inWS) * (0.01881234f);
 result += GET_MATRIX_DATA(iD, Y+20,X+17, inWS) * (0.00351649f);
 result += GET_MATRIX_DATA(iD, Y+26,X+17, inWS) * (-0.01241040f);
 result += GET_MATRIX_DATA(iD, Y+24,X+18, inWS) * (-0.00176270f);
 result += GET_MATRIX_DATA(iD, Y+26,X+18, inWS) * (-0.01364610f);
 result += GET_MATRIX_DATA(iD, Y+11,X+20, inWS) * (0.02151897f);
 result += GET_MATRIX_DATA(iD, Y+12,X+20, inWS) * (0.01273144f);
 result += GET_MATRIX_DATA(iD, Y+13,X+20, inWS) * (0.01019841f);
 result += GET_MATRIX_DATA(iD, Y+13,X+21, inWS) * (0.00612031f);
 result += GET_MATRIX_DATA(iD, Y+20,X+21, inWS) * (-0.00091414f);
 result += GET_MATRIX_DATA(iD, Y+18,X+22, inWS) * (-0.00057612f);
 result += GET_MATRIX_DATA(iD, Y+19,X+22, inWS) * (-0.03587870f);
 result += GET_MATRIX_DATA(iD, Y+23,X+24, inWS) * (0.01256504f);
 result += GET_MATRIX_DATA(iD, Y+26,X+24, inWS) * (0.03923267f);
 result += GET_MATRIX_DATA(iD, Y+22,X+25, inWS) * (0.00444355f);
 result += GET_MATRIX_DATA(iD, Y+23,X+25, inWS) * (0.01363273f);
 result += GET_MATRIX_DATA(iD, Y+24,X+25, inWS) * (0.00919381f);
 result += GET_MATRIX_DATA(iD, Y+25,X+25, inWS) * (0.00810816f);
 result += GET_MATRIX_DATA(iD, Y+13,X+26, inWS) * (-0.01089781f);
 result += GET_MATRIX_DATA(iD, Y+14,X+26, inWS) * (-0.00429632f);
 result += GET_MATRIX_DATA(iD, Y+21,X+26, inWS) * (0.00892223f);
 result += GET_MATRIX_DATA(iD, Y+22,X+26, inWS) * (0.01356452f);
 result += GET_MATRIX_DATA(iD, Y+25,X+26, inWS) * (0.02025559f);
 result += GET_MATRIX_DATA(iD, Y+22,X+27, inWS) * (0.00406188f);
 result += GET_MATRIX_DATA(iD, Y+23,X+27, inWS) * (0.00473408f);
 result += GET_MATRIX_DATA(iD, Y+24,X+27, inWS) * (0.01224033f);
 result += GET_MATRIX_DATA(iD, Y+2,X+28, inWS) * (0.00561848f);
 result += GET_MATRIX_DATA(iD, Y+2,X+29, inWS) * (0.00561364f);
 result += GET_MATRIX_DATA(iD, Y+1,X+30, inWS) * (0.00147426f);
 result += GET_MATRIX_DATA(iD, Y+2,X+30, inWS) * (0.00145660f);
 result += GET_MATRIX_DATA(iD, Y+4,X+30, inWS) * (0.00476424f);
 result += GET_MATRIX_DATA(iD, Y+30,X+32, inWS) * (-0.00823513f);
 result += GET_MATRIX_DATA(iD, Y+31,X+32, inWS) * (-0.01833530f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00684733f);
desc[26] = (result > (-51.63f));

 /***********************************************
                  Projection: 28
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+3,X+3, inWS) * (0.00714724f);
 result += GET_MATRIX_DATA(iD, Y+3,X+5, inWS) * (0.00401909f);
 result += GET_MATRIX_DATA(iD, Y+12,X+6, inWS) * (-0.01418672f);
 result += GET_MATRIX_DATA(iD, Y+21,X+6, inWS) * (0.00063429f);
 result += GET_MATRIX_DATA(iD, Y+22,X+6, inWS) * (0.02510799f);
 result += GET_MATRIX_DATA(iD, Y+10,X+8, inWS) * (-0.00529715f);
 result += GET_MATRIX_DATA(iD, Y+8,X+9, inWS) * (-0.00155759f);
 result += GET_MATRIX_DATA(iD, Y+10,X+9, inWS) * (-0.00721354f);
 result += GET_MATRIX_DATA(iD, Y+8,X+10, inWS) * (-0.02268428f);
 result += GET_MATRIX_DATA(iD, Y+6,X+11, inWS) * (-0.00458685f);
 result += GET_MATRIX_DATA(iD, Y+14,X+11, inWS) * (0.00056371f);
 result += GET_MATRIX_DATA(iD, Y+19,X+11, inWS) * (-0.01619193f);
 result += GET_MATRIX_DATA(iD, Y+13,X+12, inWS) * (0.04992311f);
 result += GET_MATRIX_DATA(iD, Y+14,X+12, inWS) * (0.00770683f);
 result += GET_MATRIX_DATA(iD, Y+21,X+12, inWS) * (-0.00824646f);
 result += GET_MATRIX_DATA(iD, Y+22,X+12, inWS) * (-0.00892291f);
 result += GET_MATRIX_DATA(iD, Y+12,X+13, inWS) * (0.09024388f);
 result += GET_MATRIX_DATA(iD, Y+22,X+13, inWS) * (-0.01495477f);
 result += GET_MATRIX_DATA(iD, Y+16,X+14, inWS) * (-0.00127637f);
 result += GET_MATRIX_DATA(iD, Y+22,X+14, inWS) * (-0.01060576f);
 result += GET_MATRIX_DATA(iD, Y+15,X+15, inWS) * (-0.01353639f);
 result += GET_MATRIX_DATA(iD, Y+19,X+15, inWS) * (0.03116076f);
 result += GET_MATRIX_DATA(iD, Y+28,X+15, inWS) * (0.01822782f);
 result += GET_MATRIX_DATA(iD, Y+30,X+15, inWS) * (0.00478740f);
 result += GET_MATRIX_DATA(iD, Y+31,X+15, inWS) * (0.00136139f);
 result += GET_MATRIX_DATA(iD, Y+15,X+16, inWS) * (-0.06124836f);
 result += GET_MATRIX_DATA(iD, Y+19,X+16, inWS) * (0.02907153f);
 result += GET_MATRIX_DATA(iD, Y+28,X+16, inWS) * (0.00112528f);
 result += GET_MATRIX_DATA(iD, Y+6,X+17, inWS) * (-0.00728673f);
 result += GET_MATRIX_DATA(iD, Y+7,X+17, inWS) * (-0.00254020f);
 result += GET_MATRIX_DATA(iD, Y+13,X+17, inWS) * (-0.00957722f);
 result += GET_MATRIX_DATA(iD, Y+14,X+17, inWS) * (-0.00242070f);
 result += GET_MATRIX_DATA(iD, Y+15,X+17, inWS) * (-0.01174473f);
 result += GET_MATRIX_DATA(iD, Y+18,X+17, inWS) * (0.01733599f);
 result += GET_MATRIX_DATA(iD, Y+28,X+17, inWS) * (0.01417952f);
 result += GET_MATRIX_DATA(iD, Y+22,X+19, inWS) * (-0.00264791f);
 result += GET_MATRIX_DATA(iD, Y+20,X+20, inWS) * (-0.00361874f);
 result += GET_MATRIX_DATA(iD, Y+19,X+21, inWS) * (-0.01803550f);
 result += GET_MATRIX_DATA(iD, Y+20,X+21, inWS) * (-0.00494727f);
 result += GET_MATRIX_DATA(iD, Y+30,X+25, inWS) * (-0.01765249f);
 result += GET_MATRIX_DATA(iD, Y+9,X+27, inWS) * (-0.00105682f);
 result += GET_MATRIX_DATA(iD, Y+12,X+27, inWS) * (-0.01343864f);
 result += GET_MATRIX_DATA(iD, Y+22,X+27, inWS) * (0.00254406f);
 result += GET_MATRIX_DATA(iD, Y+24,X+27, inWS) * (0.01805521f);
 result += GET_MATRIX_DATA(iD, Y+13,X+28, inWS) * (-0.00511341f);
 result += GET_MATRIX_DATA(iD, Y+22,X+28, inWS) * (0.00194531f);
 result += GET_MATRIX_DATA(iD, Y+31,X+28, inWS) * (-0.00872004f);
 result += GET_MATRIX_DATA(iD, Y+10,X+29, inWS) * (-0.01398125f);
 result += GET_MATRIX_DATA(iD, Y+20,X+29, inWS) * (0.00777416f);
 result += GET_MATRIX_DATA(iD, Y+21,X+29, inWS) * (0.00448072f);
 result += GET_MATRIX_DATA(iD, Y+22,X+29, inWS) * (0.01948379f);
 result += GET_MATRIX_DATA(iD, Y+24,X+29, inWS) * (0.00192448f);
 result += GET_MATRIX_DATA(iD, Y+20,X+30, inWS) * (0.01270066f);
 result += GET_MATRIX_DATA(iD, Y+22,X+30, inWS) * (0.00488468f);
 result += GET_MATRIX_DATA(iD, Y+31,X+30, inWS) * (-0.00585233f);
 result += GET_MATRIX_DATA(iD, Y+32,X+30, inWS) * (-0.02188718f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00084717f);
desc[27] = (result > ( 4.04f));

 /***********************************************
                  Projection: 29
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (0.01070677f);
 result += GET_MATRIX_DATA(iD, Y+5,X+6, inWS) * (0.00321703f);
 result += GET_MATRIX_DATA(iD, Y+2,X+8, inWS) * (0.00873639f);
 result += GET_MATRIX_DATA(iD, Y+26,X+8, inWS) * (-0.00213926f);
 result += GET_MATRIX_DATA(iD, Y+24,X+9, inWS) * (-0.02887696f);
 result += GET_MATRIX_DATA(iD, Y+25,X+9, inWS) * (-0.01873541f);
 result += GET_MATRIX_DATA(iD, Y+17,X+10, inWS) * (0.01359845f);
 result += GET_MATRIX_DATA(iD, Y+27,X+10, inWS) * (-0.00050049f);
 result += GET_MATRIX_DATA(iD, Y+28,X+10, inWS) * (-0.01012203f);
 result += GET_MATRIX_DATA(iD, Y+16,X+11, inWS) * (0.04551799f);
 result += GET_MATRIX_DATA(iD, Y+17,X+11, inWS) * (0.02819249f);
 result += GET_MATRIX_DATA(iD, Y+11,X+12, inWS) * (-0.04276160f);
 result += GET_MATRIX_DATA(iD, Y+16,X+12, inWS) * (0.00423899f);
 result += GET_MATRIX_DATA(iD, Y+5,X+13, inWS) * (-0.00424319f);
 result += GET_MATRIX_DATA(iD, Y+13,X+14, inWS) * (-0.03506015f);
 result += GET_MATRIX_DATA(iD, Y+24,X+15, inWS) * (0.01971446f);
 result += GET_MATRIX_DATA(iD, Y+25,X+15, inWS) * (0.02584927f);
 result += GET_MATRIX_DATA(iD, Y+30,X+15, inWS) * (0.01678732f);
 result += GET_MATRIX_DATA(iD, Y+9,X+16, inWS) * (0.00537900f);
 result += GET_MATRIX_DATA(iD, Y+12,X+16, inWS) * (0.00519506f);
 result += GET_MATRIX_DATA(iD, Y+19,X+16, inWS) * (-0.00310644f);
 result += GET_MATRIX_DATA(iD, Y+24,X+16, inWS) * (0.00568446f);
 result += GET_MATRIX_DATA(iD, Y+25,X+16, inWS) * (0.02743977f);
 result += GET_MATRIX_DATA(iD, Y+9,X+17, inWS) * (0.03467816f);
 result += GET_MATRIX_DATA(iD, Y+9,X+18, inWS) * (0.00206637f);
 result += GET_MATRIX_DATA(iD, Y+10,X+18, inWS) * (0.02802664f);
 result += GET_MATRIX_DATA(iD, Y+20,X+18, inWS) * (-0.01720974f);
 result += GET_MATRIX_DATA(iD, Y+2,X+19, inWS) * (-0.01231617f);
 result += GET_MATRIX_DATA(iD, Y+9,X+19, inWS) * (0.02044246f);
 result += GET_MATRIX_DATA(iD, Y+16,X+19, inWS) * (0.01863570f);
 result += GET_MATRIX_DATA(iD, Y+20,X+19, inWS) * (-0.00017617f);
 result += GET_MATRIX_DATA(iD, Y+1,X+20, inWS) * (-0.00877399f);
 result += GET_MATRIX_DATA(iD, Y+13,X+20, inWS) * (-0.00630277f);
 result += GET_MATRIX_DATA(iD, Y+14,X+20, inWS) * (-0.04641082f);
 result += GET_MATRIX_DATA(iD, Y+17,X+20, inWS) * (0.01706946f);
 result += GET_MATRIX_DATA(iD, Y+21,X+20, inWS) * (-0.01616104f);
 result += GET_MATRIX_DATA(iD, Y+23,X+20, inWS) * (-0.00328225f);
 result += GET_MATRIX_DATA(iD, Y+13,X+21, inWS) * (-0.01661073f);
 result += GET_MATRIX_DATA(iD, Y+14,X+21, inWS) * (-0.06718326f);
 result += GET_MATRIX_DATA(iD, Y+18,X+21, inWS) * (0.03448250f);
 result += GET_MATRIX_DATA(iD, Y+30,X+21, inWS) * (-0.00965598f);
 result += GET_MATRIX_DATA(iD, Y+32,X+21, inWS) * (-0.01906617f);
 result += GET_MATRIX_DATA(iD, Y+17,X+22, inWS) * (0.01669698f);
 result += GET_MATRIX_DATA(iD, Y+18,X+22, inWS) * (0.02352728f);
 result += GET_MATRIX_DATA(iD, Y+7,X+23, inWS) * (0.00847745f);
 result += GET_MATRIX_DATA(iD, Y+19,X+23, inWS) * (0.01452677f);
 result += GET_MATRIX_DATA(iD, Y+31,X+23, inWS) * (-0.00132927f);
 result += GET_MATRIX_DATA(iD, Y+7,X+25, inWS) * (0.00151153f);
 result += GET_MATRIX_DATA(iD, Y+7,X+26, inWS) * (0.00587425f);
 result += GET_MATRIX_DATA(iD, Y+14,X+26, inWS) * (-0.00126712f);
 result += GET_MATRIX_DATA(iD, Y+26,X+27, inWS) * (0.00607182f);
 result += GET_MATRIX_DATA(iD, Y+26,X+28, inWS) * (0.00207796f);
 result += GET_MATRIX_DATA(iD, Y+15,X+30, inWS) * (-0.00746446f);
 result += GET_MATRIX_DATA(iD, Y+26,X+30, inWS) * (0.00654237f);
 result += GET_MATRIX_DATA(iD, Y+28,X+30, inWS) * (0.00370861f);
 result += GET_MATRIX_DATA(iD, Y+15,X+31, inWS) * (-0.02824838f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00107613f);
desc[28] = (result > (79.59f));

 /***********************************************
                  Projection: 30
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+3,X+1, inWS) * (0.00227777f);
 result += GET_MATRIX_DATA(iD, Y+1,X+2, inWS) * (0.01057954f);
 result += GET_MATRIX_DATA(iD, Y+3,X+2, inWS) * (0.02102098f);
 result += GET_MATRIX_DATA(iD, Y+3,X+3, inWS) * (0.02409609f);
 result += GET_MATRIX_DATA(iD, Y+24,X+3, inWS) * (-0.00621433f);
 result += GET_MATRIX_DATA(iD, Y+17,X+5, inWS) * (0.01354394f);
 result += GET_MATRIX_DATA(iD, Y+24,X+5, inWS) * (-0.00527748f);
 result += GET_MATRIX_DATA(iD, Y+8,X+6, inWS) * (-0.01211030f);
 result += GET_MATRIX_DATA(iD, Y+10,X+7, inWS) * (-0.03089298f);
 result += GET_MATRIX_DATA(iD, Y+17,X+7, inWS) * (0.01241876f);
 result += GET_MATRIX_DATA(iD, Y+6,X+8, inWS) * (-0.00021309f);
 result += GET_MATRIX_DATA(iD, Y+7,X+8, inWS) * (-0.01300596f);
 result += GET_MATRIX_DATA(iD, Y+8,X+8, inWS) * (-0.01825499f);
 result += GET_MATRIX_DATA(iD, Y+15,X+8, inWS) * (0.00441262f);
 result += GET_MATRIX_DATA(iD, Y+5,X+9, inWS) * (-0.02661148f);
 result += GET_MATRIX_DATA(iD, Y+6,X+9, inWS) * (-0.00319017f);
 result += GET_MATRIX_DATA(iD, Y+15,X+9, inWS) * (0.00784502f);
 result += GET_MATRIX_DATA(iD, Y+16,X+9, inWS) * (0.00870659f);
 result += GET_MATRIX_DATA(iD, Y+17,X+9, inWS) * (0.00416173f);
 result += GET_MATRIX_DATA(iD, Y+2,X+10, inWS) * (-0.00302104f);
 result += GET_MATRIX_DATA(iD, Y+3,X+10, inWS) * (-0.00097553f);
 result += GET_MATRIX_DATA(iD, Y+4,X+10, inWS) * (-0.01475694f);
 result += GET_MATRIX_DATA(iD, Y+15,X+10, inWS) * (0.00963283f);
 result += GET_MATRIX_DATA(iD, Y+14,X+11, inWS) * (0.02883831f);
 result += GET_MATRIX_DATA(iD, Y+1,X+12, inWS) * (-0.00536472f);
 result += GET_MATRIX_DATA(iD, Y+19,X+12, inWS) * (-0.00700165f);
 result += GET_MATRIX_DATA(iD, Y+20,X+12, inWS) * (-0.01631597f);
 result += GET_MATRIX_DATA(iD, Y+19,X+13, inWS) * (-0.01553821f);
 result += GET_MATRIX_DATA(iD, Y+8,X+14, inWS) * (0.02010527f);
 result += GET_MATRIX_DATA(iD, Y+8,X+15, inWS) * (0.00626201f);
 result += GET_MATRIX_DATA(iD, Y+26,X+15, inWS) * (0.00740225f);
 result += GET_MATRIX_DATA(iD, Y+7,X+16, inWS) * (0.01331765f);
 result += GET_MATRIX_DATA(iD, Y+8,X+16, inWS) * (0.03178506f);
 result += GET_MATRIX_DATA(iD, Y+12,X+16, inWS) * (-0.03225718f);
 result += GET_MATRIX_DATA(iD, Y+8,X+17, inWS) * (0.01700234f);
 result += GET_MATRIX_DATA(iD, Y+11,X+17, inWS) * (-0.01618030f);
 result += GET_MATRIX_DATA(iD, Y+22,X+17, inWS) * (0.01263102f);
 result += GET_MATRIX_DATA(iD, Y+31,X+17, inWS) * (-0.00857377f);
 result += GET_MATRIX_DATA(iD, Y+11,X+18, inWS) * (-0.00057118f);
 result += GET_MATRIX_DATA(iD, Y+12,X+18, inWS) * (-0.00178769f);
 result += GET_MATRIX_DATA(iD, Y+15,X+18, inWS) * (0.02160969f);
 result += GET_MATRIX_DATA(iD, Y+22,X+18, inWS) * (0.01038029f);
 result += GET_MATRIX_DATA(iD, Y+11,X+19, inWS) * (-0.04801443f);
 result += GET_MATRIX_DATA(iD, Y+22,X+19, inWS) * (0.00320583f);
 result += GET_MATRIX_DATA(iD, Y+30,X+19, inWS) * (-0.00197627f);
 result += GET_MATRIX_DATA(iD, Y+15,X+20, inWS) * (0.01012582f);
 result += GET_MATRIX_DATA(iD, Y+19,X+20, inWS) * (-0.03106581f);
 result += GET_MATRIX_DATA(iD, Y+30,X+20, inWS) * (-0.00205132f);
 result += GET_MATRIX_DATA(iD, Y+24,X+21, inWS) * (0.01058858f);
 result += GET_MATRIX_DATA(iD, Y+3,X+23, inWS) * (-0.00292018f);
 result += GET_MATRIX_DATA(iD, Y+4,X+23, inWS) * (-0.00361334f);
 result += GET_MATRIX_DATA(iD, Y+5,X+23, inWS) * (-0.00184230f);
 result += GET_MATRIX_DATA(iD, Y+22,X+23, inWS) * (0.00263153f);
 result += GET_MATRIX_DATA(iD, Y+23,X+23, inWS) * (0.00741852f);
 result += GET_MATRIX_DATA(iD, Y+11,X+27, inWS) * (0.00180050f);
 result += GET_MATRIX_DATA(iD, Y+19,X+29, inWS) * (-0.00806244f);
 result += GET_MATRIX_DATA(iD, Y+21,X+29, inWS) * (-0.00031341f);
 result += GET_MATRIX_DATA(iD, Y+20,X+30, inWS) * (-0.00606769f);
 result += GET_MATRIX_DATA(iD, Y+16,X+31, inWS) * (-0.00661586f);
 result += GET_MATRIX_DATA(iD, Y+17,X+31, inWS) * (-0.00274089f);
 result += GET_MATRIX_DATA(iD, Y+20,X+31, inWS) * (-0.01937966f);
 result += GET_MATRIX_DATA(iD, Y+30,X+31, inWS) * (0.01317872f);
 result += GET_MATRIX_DATA(iD, Y+31,X+31, inWS) * (0.00299545f);
 result += GET_MATRIX_DATA(iD, Y+30,X+32, inWS) * (0.01773691f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (0.00114575f);
desc[29] = (result > (49.18f));

 /***********************************************
                  Projection: 31
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+14,X+5, inWS) * (-0.00428443f);
 result += GET_MATRIX_DATA(iD, Y+13,X+6, inWS) * (-0.02334822f);
 result += GET_MATRIX_DATA(iD, Y+21,X+6, inWS) * (0.01554221f);
 result += GET_MATRIX_DATA(iD, Y+19,X+7, inWS) * (0.00096422f);
 result += GET_MATRIX_DATA(iD, Y+21,X+7, inWS) * (0.01643749f);
 result += GET_MATRIX_DATA(iD, Y+12,X+8, inWS) * (-0.01120805f);
 result += GET_MATRIX_DATA(iD, Y+19,X+8, inWS) * (0.01753613f);
 result += GET_MATRIX_DATA(iD, Y+12,X+9, inWS) * (-0.00072254f);
 result += GET_MATRIX_DATA(iD, Y+13,X+9, inWS) * (-0.00859184f);
 result += GET_MATRIX_DATA(iD, Y+14,X+9, inWS) * (-0.00397050f);
 result += GET_MATRIX_DATA(iD, Y+19,X+9, inWS) * (0.02132196f);
 result += GET_MATRIX_DATA(iD, Y+21,X+9, inWS) * (0.00503141f);
 result += GET_MATRIX_DATA(iD, Y+22,X+9, inWS) * (0.00270490f);
 result += GET_MATRIX_DATA(iD, Y+27,X+9, inWS) * (-0.01095149f);
 result += GET_MATRIX_DATA(iD, Y+13,X+10, inWS) * (-0.00842014f);
 result += GET_MATRIX_DATA(iD, Y+2,X+11, inWS) * (-0.01125056f);
 result += GET_MATRIX_DATA(iD, Y+14,X+11, inWS) * (-0.04059970f);
 result += GET_MATRIX_DATA(iD, Y+15,X+11, inWS) * (-0.00083485f);
 result += GET_MATRIX_DATA(iD, Y+20,X+11, inWS) * (0.06019970f);
 result += GET_MATRIX_DATA(iD, Y+27,X+11, inWS) * (-0.01146334f);
 result += GET_MATRIX_DATA(iD, Y+7,X+12, inWS) * (0.01659759f);
 result += GET_MATRIX_DATA(iD, Y+8,X+12, inWS) * (0.00506546f);
 result += GET_MATRIX_DATA(iD, Y+1,X+13, inWS) * (-0.00830110f);
 result += GET_MATRIX_DATA(iD, Y+2,X+13, inWS) * (-0.01512850f);
 result += GET_MATRIX_DATA(iD, Y+25,X+13, inWS) * (-0.00368036f);
 result += GET_MATRIX_DATA(iD, Y+26,X+13, inWS) * (-0.03860396f);
 result += GET_MATRIX_DATA(iD, Y+1,X+14, inWS) * (-0.01090421f);
 result += GET_MATRIX_DATA(iD, Y+8,X+14, inWS) * (0.02960920f);
 result += GET_MATRIX_DATA(iD, Y+9,X+14, inWS) * (0.02791716f);
 result += GET_MATRIX_DATA(iD, Y+25,X+14, inWS) * (-0.01078672f);
 result += GET_MATRIX_DATA(iD, Y+8,X+15, inWS) * (0.05067817f);
 result += GET_MATRIX_DATA(iD, Y+16,X+15, inWS) * (0.01469685f);
 result += GET_MATRIX_DATA(iD, Y+18,X+15, inWS) * (-0.04229109f);
 result += GET_MATRIX_DATA(iD, Y+25,X+15, inWS) * (-0.02542690f);
 result += GET_MATRIX_DATA(iD, Y+31,X+15, inWS) * (0.00743726f);
 result += GET_MATRIX_DATA(iD, Y+32,X+15, inWS) * (0.02688940f);
 result += GET_MATRIX_DATA(iD, Y+1,X+16, inWS) * (-0.02893727f);
 result += GET_MATRIX_DATA(iD, Y+11,X+16, inWS) * (-0.00732222f);
 result += GET_MATRIX_DATA(iD, Y+12,X+16, inWS) * (-0.00301955f);
 result += GET_MATRIX_DATA(iD, Y+16,X+16, inWS) * (0.02973268f);
 result += GET_MATRIX_DATA(iD, Y+32,X+16, inWS) * (0.01874856f);
 result += GET_MATRIX_DATA(iD, Y+18,X+17, inWS) * (-0.00035814f);
 result += GET_MATRIX_DATA(iD, Y+30,X+17, inWS) * (0.00215555f);
 result += GET_MATRIX_DATA(iD, Y+32,X+17, inWS) * (0.02069455f);
 result += GET_MATRIX_DATA(iD, Y+1,X+18, inWS) * (-0.01971264f);
 result += GET_MATRIX_DATA(iD, Y+2,X+18, inWS) * (-0.01876831f);
 result += GET_MATRIX_DATA(iD, Y+11,X+18, inWS) * (-0.01851555f);
 result += GET_MATRIX_DATA(iD, Y+30,X+18, inWS) * (0.01347780f);
 result += GET_MATRIX_DATA(iD, Y+2,X+19, inWS) * (-0.01217582f);
 result += GET_MATRIX_DATA(iD, Y+11,X+19, inWS) * (-0.02404988f);
 result += GET_MATRIX_DATA(iD, Y+22,X+19, inWS) * (0.02951939f);
 result += GET_MATRIX_DATA(iD, Y+32,X+19, inWS) * (0.00637488f);
 result += GET_MATRIX_DATA(iD, Y+26,X+20, inWS) * (-0.00491490f);
 result += GET_MATRIX_DATA(iD, Y+32,X+20, inWS) * (0.00811210f);
 result += GET_MATRIX_DATA(iD, Y+14,X+21, inWS) * (0.00376611f);
 result += GET_MATRIX_DATA(iD, Y+26,X+22, inWS) * (-0.00692990f);
 result += GET_MATRIX_DATA(iD, Y+5,X+24, inWS) * (0.00704987f);
 result += GET_MATRIX_DATA(iD, Y+6,X+24, inWS) * (0.00551815f);
 result += GET_MATRIX_DATA(iD, Y+8,X+24, inWS) * (0.00332611f);
 result += GET_MATRIX_DATA(iD, Y+27,X+24, inWS) * (-0.02599153f);
 result += GET_MATRIX_DATA(iD, Y+4,X+25, inWS) * (0.00076713f);
 result += GET_MATRIX_DATA(iD, Y+2,X+26, inWS) * (0.00168109f);
 result += GET_MATRIX_DATA(iD, Y+3,X+26, inWS) * (0.01394147f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00055366f);
desc[30] = (result > (-0.58f));

 /***********************************************
                  Projection: 32
 ************************************************/
 result = 0.f;
 result += GET_MATRIX_DATA(iD, Y+30,X+2, inWS) * (0.00607372f);
 result += GET_MATRIX_DATA(iD, Y+22,X+3, inWS) * (-0.02550222f);
 result += GET_MATRIX_DATA(iD, Y+30,X+3, inWS) * (0.01289829f);
 result += GET_MATRIX_DATA(iD, Y+31,X+3, inWS) * (0.00606745f);
 result += GET_MATRIX_DATA(iD, Y+15,X+5, inWS) * (0.00828729f);
 result += GET_MATRIX_DATA(iD, Y+23,X+5, inWS) * (-0.00869528f);
 result += GET_MATRIX_DATA(iD, Y+31,X+5, inWS) * (0.01704377f);
 result += GET_MATRIX_DATA(iD, Y+16,X+7, inWS) * (0.00705650f);
 result += GET_MATRIX_DATA(iD, Y+24,X+7, inWS) * (-0.02863963f);
 result += GET_MATRIX_DATA(iD, Y+17,X+9, inWS) * (0.00685146f);
 result += GET_MATRIX_DATA(iD, Y+18,X+9, inWS) * (0.01090355f);
 result += GET_MATRIX_DATA(iD, Y+25,X+9, inWS) * (-0.01507082f);
 result += GET_MATRIX_DATA(iD, Y+8,X+10, inWS) * (-0.00090832f);
 result += GET_MATRIX_DATA(iD, Y+18,X+10, inWS) * (0.03630597f);
 result += GET_MATRIX_DATA(iD, Y+19,X+10, inWS) * (0.02242218f);
 result += GET_MATRIX_DATA(iD, Y+12,X+11, inWS) * (-0.02398435f);
 result += GET_MATRIX_DATA(iD, Y+13,X+11, inWS) * (-0.02324572f);
 result += GET_MATRIX_DATA(iD, Y+20,X+11, inWS) * (0.00671353f);
 result += GET_MATRIX_DATA(iD, Y+14,X+12, inWS) * (-0.02455580f);
 result += GET_MATRIX_DATA(iD, Y+15,X+12, inWS) * (-0.01442296f);
 result += GET_MATRIX_DATA(iD, Y+19,X+12, inWS) * (0.00178093f);
 result += GET_MATRIX_DATA(iD, Y+3,X+13, inWS) * (0.00575224f);
 result += GET_MATRIX_DATA(iD, Y+32,X+14, inWS) * (-0.00642457f);
 result += GET_MATRIX_DATA(iD, Y+2,X+15, inWS) * (0.00817225f);
 result += GET_MATRIX_DATA(iD, Y+3,X+15, inWS) * (0.00716109f);
 result += GET_MATRIX_DATA(iD, Y+11,X+15, inWS) * (0.06087673f);
 result += GET_MATRIX_DATA(iD, Y+12,X+15, inWS) * (0.00355298f);
 result += GET_MATRIX_DATA(iD, Y+16,X+15, inWS) * (-0.00889138f);
 result += GET_MATRIX_DATA(iD, Y+22,X+15, inWS) * (-0.01068958f);
 result += GET_MATRIX_DATA(iD, Y+30,X+15, inWS) * (-0.00283261f);
 result += GET_MATRIX_DATA(iD, Y+32,X+15, inWS) * (-0.00408522f);
 result += GET_MATRIX_DATA(iD, Y+18,X+16, inWS) * (0.00377506f);
 result += GET_MATRIX_DATA(iD, Y+22,X+16, inWS) * (-0.01641793f);
 result += GET_MATRIX_DATA(iD, Y+25,X+16, inWS) * (0.00829984f);
 result += GET_MATRIX_DATA(iD, Y+18,X+17, inWS) * (0.01401543f);
 result += GET_MATRIX_DATA(iD, Y+25,X+17, inWS) * (0.01800033f);
 result += GET_MATRIX_DATA(iD, Y+26,X+17, inWS) * (0.01226503f);
 result += GET_MATRIX_DATA(iD, Y+13,X+18, inWS) * (-0.01518862f);
 result += GET_MATRIX_DATA(iD, Y+18,X+18, inWS) * (0.01766956f);
 result += GET_MATRIX_DATA(iD, Y+5,X+19, inWS) * (-0.00294398f);
 result += GET_MATRIX_DATA(iD, Y+7,X+19, inWS) * (-0.00457603f);
 result += GET_MATRIX_DATA(iD, Y+8,X+19, inWS) * (-0.00554052f);
 result += GET_MATRIX_DATA(iD, Y+12,X+19, inWS) * (-0.00793670f);
 result += GET_MATRIX_DATA(iD, Y+25,X+19, inWS) * (0.02467540f);
 result += GET_MATRIX_DATA(iD, Y+26,X+19, inWS) * (0.03079909f);
 result += GET_MATRIX_DATA(iD, Y+10,X+20, inWS) * (-0.00807469f);
 result += GET_MATRIX_DATA(iD, Y+15,X+21, inWS) * (0.00643537f);
 result += GET_MATRIX_DATA(iD, Y+20,X+21, inWS) * (-0.03399521f);
 result += GET_MATRIX_DATA(iD, Y+20,X+22, inWS) * (-0.02313247f);
 result += GET_MATRIX_DATA(iD, Y+27,X+22, inWS) * (0.00134224f);
 result += GET_MATRIX_DATA(iD, Y+13,X+23, inWS) * (0.00329653f);
 result += GET_MATRIX_DATA(iD, Y+20,X+23, inWS) * (-0.00857958f);
 result += GET_MATRIX_DATA(iD, Y+21,X+23, inWS) * (-0.02712394f);
 result += GET_MATRIX_DATA(iD, Y+12,X+24, inWS) * (0.01050558f);
 result += GET_MATRIX_DATA(iD, Y+13,X+24, inWS) * (0.02498465f);
 result += GET_MATRIX_DATA(iD, Y+21,X+24, inWS) * (-0.00878718f);
 result += GET_MATRIX_DATA(iD, Y+9,X+25, inWS) * (0.00065978f);
 result += GET_MATRIX_DATA(iD, Y+21,X+25, inWS) * (-0.00615120f);
result += (GET_MATRIX_DATA(iD, kpt.pt.y, kpt.pt.x, inWS)) * (-0.00010379f);
desc[31] = (result > (72.97f));
}
